/**************************************************************

aflat.c (C-Munipack project)
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <fitsio.h>
#include <math.h>
#include <time.h>

#include "comfun.h"
#include "fitsfun.h"
#include "robmean.h"
#include "imglist.h"
#include "inifile.h"
#include "aflat.h"

int   aflat_maxsky;          // Max. number of pixels used for sky level coputation
double aflat_level;          // Mean output level (required)
int   aflat_bitpix;          // No. of bits per pixel
int   aflat_nx;              // Width of the frame
int   aflat_ny;              // Height of the frame
double aflat_minvalue;       // Lowest good datum
double aflat_maxvalue;       // Highest good datum
char  aflat_filter[MAXLINE]; // Name of filter
double aflat_exp;            // Exposure duration
double aflat_temp;           // Exposure duration
char  aflat_date[MAXLINE];   // Date observation of first frame
char  aflat_time[MAXLINE];   // Time observation of first frame
tImgList aflat_list = NULL;  // List of frames
int aflat_out_bitpix;		 // Output data format

int aflat_init(const char *optfile)
// Initializes variables and image list
{
  inifile *f;
  char msg[MAXLINE];

  aflat_nx      = 0;
  aflat_ny      = 0;
  aflat_list    = ImgList_Create(0, 0); 
  aflat_date[0] = 0;
  aflat_time[0] = 0;

  /* Read parameter file */
  if ((f=iniopen(optfile, OPTFILE))==NULL) {
	if (optfile) sprintf(msg,"Cannot read parameter file '%s'", optfile);
		    else sprintf(msg,"Cannot read parameter file '%s'", OPTFILE);
	PrintOut(0, msg);
  }
  aflat_maxsky   = inigkyl(f, "MAXSKY", 10000);
  aflat_level    = inigkyd(f, "LEVEL",  1000.0);
  aflat_bitpix 	 = inigkyl(f, "BITPIX", 0);
  iniclos(f);

  /* Print parameters */
  PrintOut(1, "Configuration parameters:");
  sprintf(msg, "MAXSKY: %d", aflat_maxsky); 	
  PrintOut(1, msg);
  sprintf(msg, "LEVEL : %.2f", aflat_level); 
  PrintOut(1, msg);
  sprintf(msg, "BITPIX: %d", aflat_bitpix); 
  PrintOut(1, msg);
  PrintOut(1, "-------------------------------");
    
  /* Check parameters */
  if (aflat_maxsky<=0) {
     PrintOut(0, "Invalid value of maxsky. Must be greater than zero.");
	 return MPK_INVALID_PAR;
  }
  if (aflat_level<=0) {
     PrintOut(0, "Invalid value of level. Must be greater than zero.");
     return MPK_INVALID_PAR;
  }
  
  /* Normal return */
  return 0;
}

int aflat_clean(void)
// Cleans the image list
{
	double *ccd;
	
	/* Free allocated images and clear the image list */
    if (ImgList_First(aflat_list)) {
        do {
	        ccd = ImgList_GetImage(aflat_list);
            if (ccd) free(ccd);
            ImgList_SetImage(aflat_list, NULL);
        } while (ImgList_Next(aflat_list));
    }
	ImgList_Destroy(aflat_list);
	aflat_list = NULL;
	
	/* Normal return */
	return 0;
}

int aflat_read(const char *filename)
// Reads one image and puts it into the list
{
	FILE *f;
	fitsfile *fits;
	int istat = 0, anyf, nx, ny, i, bitpix;
    double skymed, skysig, r, *ccd;
    double exptime, temp;
    char buf[MAXLINE], filter[MAXLINE], date[MAXLINE], time[MAXLINE];

    /* Check source file */
    if (!filename) {
        PrintOut(0, "Missing name of the source image file.");
 	    return MPK_INVALID_PAR;
    }
    if (strlen(filename)==0) {
        PrintOut(0, "Missing name of the source image file.");
 	    return MPK_INVALID_PAR;
    }
    f = fopen(filename, "r");
    if (f==NULL) {
	    sprintf(buf, "Source file '%s' doesn't exists.", filename);
	    PrintOut(1, buf);
        return MPK_CANNOT_OPEN_SRC;
    }
    fclose(f);

    sprintf(buf,"Image file   : %s",filename);
    PrintOut(0, buf);

    /* open input file */
    ffdkopn(&fits,filename,READONLY,&istat);
    if (istat!=0) {
	    ffclos(fits, &istat);
	    return istat;
    }

    /* Read parameters */
    fits_gpar(fits, &nx, &ny, &bitpix);
	fits_gexp(fits, &exptime);
	fits_gfilter(fits, filter);
    fits_getdatetime(fits, date, time);
    fits_gtemp(fits, &temp);
    
  	sprintf(buf, "Image size   : %d x %d pixels", nx, ny); 
  	PrintOut(1, buf);
  	sprintf(buf, "Date & time  : %s %s UT" , date, time); 
  	PrintOut(1, buf);
	sprintf(buf, "Exposure     : %.2f s", exptime); 
  	PrintOut(1, buf);
    sprintf(buf, "Filter       : %s", filter); 
  	PrintOut(1, buf);
    sprintf(buf, "Temperature  : %.2f deg. C", temp); 
  	PrintOut(1, buf);

    if ((nx==0)||(ny==0)) {
	    /* check image size */
	    ffclos(fits,&istat);
	    return MPK_INVALID_SIZE;
    } else
    if ((aflat_nx==0)&&(aflat_ny==0)) {
        /* first image, sets the variables */
        aflat_nx = nx;
        aflat_ny = ny;
        aflat_exp = exptime;
        aflat_temp = temp;
        strncpy(aflat_filter, filter, MAXLINE);
        strncpy(aflat_date, date, MAXLINE);
        strncpy(aflat_time, time, MAXLINE);
        if (aflat_bitpix==0) {
	        /* Use data format from the first file */
	        ffgiet(fits, &aflat_out_bitpix, &istat);
    	} else {
   	        /* Use format from the configuration file */
	        aflat_out_bitpix = aflat_bitpix;
        }
        /* Do not use negative value */
        aflat_minvalue = 0.0;
		switch (aflat_out_bitpix) 
		{
		case BYTE_IMG:  
		case SBYTE_IMG:  
		case SHORT_IMG:
		case USHORT_IMG:
		    aflat_out_bitpix = USHORT_IMG;
			aflat_maxvalue = 65535.0;
			break;
		case LONG_IMG:
		case ULONG_IMG:
		    aflat_out_bitpix = ULONG_IMG;
			aflat_maxvalue = 4294967296.0;
			break;
		default:
			aflat_maxvalue = 65535.0;
		}
		sprintf(buf, "Data format  : %d", aflat_out_bitpix); 
		PrintOut(1, buf);
		sprintf(buf, "Min. value   : %.1f", aflat_minvalue); 
		PrintOut(1, buf);
		sprintf(buf, "Max. value   : %.1f", aflat_maxvalue); 
		PrintOut(1, buf);
    } else
    if ((nx!=aflat_nx)||(ny!=aflat_ny)) {
	    /* All input images must have the same size */
	    ffclos(fits,&istat);
	    return AFL_DIFFERENET_SIZE;
    }

    /* Read image data */
    ccd = malloc(nx*ny*sizeof(double));
    ffg2dd(fits,1,0.0,nx,nx,ny,ccd,&anyf,&istat);
    if (istat!=0) {
	    ffclos(fits,&istat);
	    return istat;
    }
    ffclos(fits,&istat);

    /* Image median computation */
    cmpack_fastskyD(nx, ny, ccd, aflat_maxsky, aflat_minvalue, aflat_maxvalue, &skymed, &skysig);
    sprintf(buf,"Median       : %.4f",skymed); 
  	PrintOut(0, buf);
	sprintf(buf,"Skysig       : %.4f",skysig);
  	PrintOut(0, buf);
    PrintOut(0, "-------------------------------");

    /* Normalization to aflat_level */
    r = aflat_level / skymed;
    for (i=0;i<nx*ny;i++) ccd[i]=ccd[i]*r;

    /* Add this image to image list */
    if (!ImgList_Add(aflat_list, ccd)) {
        free(ccd);
        return MPK_MEMORY;
    }

    /* Normal return */
    return 0;
}

int aflat_write(const char *outfile, double *skymed, double *skysig)
// Computes the output image and writes it to the file
{
  fitsfile *fits;
  int istat = 0, n, i, j, k, nx, ny;
  int underflow = 0, overflow = 0;
  double a, b, *fbuf, *flat, **ccds;
  long naxes[2];
  char buf[MAXLINE];
  time_t tsec;
  struct tm *actim;

  n  = ImgList_Count(aflat_list);
  if (n<=0) return MPK_NO_INPUT_FILES;

  /* makes array of images for faster access to data */
  ccds = malloc(n*sizeof(double*));
  ImgList_First(aflat_list);
  for (k=0; k<n; k++) {
	  ccds[k] = ImgList_GetImage(aflat_list);
      ImgList_Next(aflat_list);
  }

  /* Computation of median of values for each pixel */
  nx = aflat_nx;
  ny = aflat_ny;
  flat = malloc(nx*ny*sizeof(double));
  fbuf = malloc(n*sizeof(double));
  for (j=0;j<ny;j++) {
      for (i=0;i<nx;i++) {
	      /* Select pixel values from all frames */
          for (k=0;k<n;k++)
	          fbuf[k]=ccds[k][i+j*nx];
          /* Compute their median */
          cmpack_robustmeanD(n, fbuf, &a, &b);
          /* Range checking */
          if (a<aflat_minvalue) {
	          underflow = 1;
              a = aflat_minvalue;
          } else
          if (a>aflat_maxvalue) {
              overflow = 1;
              a = aflat_maxvalue;
          }
          /* Store the median into output frame */
          flat[i+j*nx] = a;
      }
  }
  free(fbuf);
  free(ccds);

  /* Compute the media of output frame */
  cmpack_fastskyD(nx, ny, flat, aflat_maxsky, aflat_minvalue, aflat_maxvalue, &a, &b);
  *skymed = a; *skysig = b;

  /* Open output file */
  sprintf(buf,"Output file  : %s", outfile);
  PrintOut(0, buf);
  sprintf(buf,"Final median : %.1f",*skymed);
  PrintOut(0, buf);
  sprintf(buf,"Final skysig : %.1f",*skysig);
  PrintOut(0, buf);
  
  remove(outfile);
  ffdkinit(&fits,outfile,&istat);

  /* Actual date */
  tsec = time(&tsec);
  actim = localtime(&tsec);

  /* Write header information */
  naxes[0]=nx; naxes[1]=ny;
  ffphps(fits,aflat_out_bitpix,2,naxes,&istat);
  ffpkys(fits,"OBJECT","Flat-field"," ",&istat);
  ffpkyg(fits,"EXPTIME",aflat_exp,2,"EXPOSURE IN SECONDS",&istat);
  if (aflat_temp<999.0)
  	ffpkyg(fits,"TEMPERAT",aflat_temp,2,"TEMPERATURE IN DEGREES C",&istat);
  ffpkyg(fits,"SKYMOD",*skymed,2,"MEAN LEVEL",&istat);
  ffpkyg(fits,"SKYSIG",*skysig,2,"FLAT-FIELD NOISE",&istat);
  ffpkys(fits,"FILTER",aflat_filter,"OPTICAL FILTER NAME",&istat);
  cmpack_datetostr(actim->tm_year+1900, actim->tm_mon+1, actim->tm_mday, buf);
  ffpkys(fits,"DATE", buf, "DATE OF CREATION",&istat);
  sprintf(buf,"%s (%s) %s", PROGNAME, PKGNAME, VERSION);
  ffpkys(fits,"ORIGIN",buf,"",&istat);
  ffukys(fits,"DATE-OBS", aflat_date, "UT DATE OF START", &istat);
  ffukys(fits,"TIME-OBS", aflat_time, "UT TIME OF START", &istat);

  /* Write image data */
  ffp2dd(fits,1,nx,nx,ny,flat,&istat);
  if (istat!=0) return istat;

  /* Close the output file */
  ffclos(fits,&istat);
  if (istat!=0) return istat;

  /* Returns range check error code or 0 */
  if (overflow) {
     sprintf(buf,"An overflow has been occurred during computation (max.=%.1f).", aflat_maxvalue);
     PrintOut(1, buf);
     //return MPK_DATA_OVERFLOW;
  }
  if (underflow) {
   	 sprintf(buf,"An underflow has been occurred during computation (min.=%.1f).", aflat_minvalue);
     PrintOut(1, buf);
     //return MPK_DATA_UNDERFLOW;
  }
  return 0;
}
