/**************************************************************

aflat.c (C-Munipack project)
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <fitsio.h>
#include <math.h>
#include <time.h>

#include "comfun.h"
#include "fitsfun.h"
#include "robmean.h"
#include "inifile.h"
#include "kombine.h"

int    kombine_nx;              // Width of the frame
int    kombine_ny;              // Height of the frame
int    kombine_bitpix;			// No. of bits per pixel
double kombine_minvalue;        // Lowest good datum
double kombine_maxvalue;        // Highest good datum
fitsfile *kombine_outf;			// Output file handle
int    kombine_ostat;			// Output file status
double *kombine_data;			// Accumulation frame
char   *kombine_stat;			// Pixel status flags
double kombine_jd;				// Accumulated JD
double kombine_exptime;			// Accumulated exposure time
int    kombine_nframe;			// Number of source frames
char   *kombine_fname;			// Output file name
int    kombine_out_bitpix;      // Output data format

#define I(x,y)	ccd[x+nx*y]
#define D(x,y)	kombine_data[x+nx*y]
#define S(x,y)	kombine_stat[x+nx*y]

int kombine_init(const char *optfile)
// Initializes variables and image list
{
	kombine_nx        = 0;
	kombine_ny        = 0;
	kombine_minvalue  = 0.0;
	kombine_maxvalue  = 65535.0;
	kombine_outf	  = NULL;
	kombine_stat      = NULL;
	kombine_data	  = NULL;
	kombine_ostat 	  = 0;
	kombine_jd        = 0.0;
	kombine_exptime   = 0.0;
	kombine_nframe 	  = 0;
	kombine_fname     = NULL;
	kombine_bitpix    = 0;
	
	/* Normal return */
	return 0;
}

int kombine_open(const char *outfile)
// Reads first image and opens destination file
{
    /* first image, sets the variables */
    kombine_nx = kombine_ny = 0;
    kombine_jd = kombine_exptime = 0;
  	kombine_nframe = 0;

    /* Open destination file */
    kombine_ostat = 0;
 	remove(outfile);
 	ffdkinit(&kombine_outf,outfile,&kombine_ostat);
 	if (kombine_ostat!=0) {
	 	kombine_outf = NULL;
	 	return MPK_CANNOT_OPEN_OUT;
 	}
 	
 	/* Store output file name */
    kombine_fname = malloc((strlen(outfile)+1)*sizeof(char));
    if (!kombine_fname) {
	    ffclos(kombine_outf,&kombine_ostat);
	    kombine_outf = NULL;
	    return MPK_MEMORY;
    }
    strcpy(kombine_fname, outfile);

    /* Normal return */
    return 0;
}

int kombine(const char *infile)
// Reads first image and opens destination file
{
	FILE *f;
	fitsfile *fits;
	int istat = 0, anyf, nx, ny, i, j, ii, jj, bitpix, offsetx, offsety;
    double *ccd;
    double exptime, val, dx, dy;
    char buf[MAXLINE], filter[MAXLINE], datestr[MAXLINE], timestr[MAXLINE];
    char *matfile, *aux, *end;

    /* Check source file */
    if (!infile) {
        PrintOut(0, "Missing name of the source image file.");
 	    return MPK_INVALID_PAR;
    }
    if (strlen(infile)==0) {
        PrintOut(0, "Missing name of the source image file.");
 	    return MPK_INVALID_PAR;
    }
    f = fopen(infile, "r");
    if (f==NULL) {
	    sprintf(buf, "Source file '%s' doesn't exists.", infile);
	    PrintOut(1, buf);
        return MPK_CANNOT_OPEN_SRC;
    }
    fclose(f);

    PrintOut(0, "----------------------------------------------");
    sprintf(buf,"Image file   : %s",infile);
    PrintOut(0, buf);
    
    /* check matched photometry file */
    matfile = (char*) malloc(strlen(infile)+5);
    if (!matfile) return MPK_MEMORY;
    strcpy(matfile, infile);
    aux = strrchr(matfile, '.');
    if (aux) strcpy(aux, ".mat");
    
    sprintf(buf,"MAT file     : %s", matfile);
    PrintOut(0, buf);
    f = fopen(matfile, "r");
    if (f==NULL) {
	    sprintf(buf, "Photometry file '%s' doesn't exists.", matfile);
	    PrintOut(1, buf);
        return KMB_MAT_NOT_FOUND;
    }
  	fgets(buf,MAXLINE,f);
  	if ((!strstr(buf," NL"))||(!strstr(buf,"JD")))
      	return MPK_UNKNOWN_FORMAT;

  	/* Read offsets */
  	fgets(buf,MAXLINE,f);
  	if (strlen(buf)>=132) {
	  	aux = buf+115;
	  	dx = strtod(aux, &end);
	  	offsetx = (int)(dx>=0.0 ? dx+0.5 : dx-0.5);
      	if (end==aux) offsetx = 0;
      	aux = buf+124;
      	dy = strtod(aux, &end);
	  	offsety = (int)(dy>=0.0 ? dy+0.5 : dy-0.5);
      	if (end==aux) offsety = 0;
  	} else {
	  	offsetx = 0;
	  	offsety = 0;
  	}
    fclose(f);

    /* open input file */
    ffdkopn(&fits,infile,READONLY,&istat);
    if (istat!=0) {
	    ffclos(fits, &istat);
	    return istat;
    }

    /* Read parameters */
    fits_gpar(fits, &nx, &ny, &bitpix);
	fits_gexp(fits, &exptime);
	fits_gfilter(fits, filter);
    fits_getdatetime(fits, datestr, timestr);
    
  	sprintf(buf, "Image size   : %d x %d pixels", nx, ny); 
  	PrintOut(0, buf);
  	sprintf(buf, "Date & time  : %s %s UT" , datestr, timestr); 
  	PrintOut(0, buf);
	sprintf(buf, "Exposure     : %.2f s", exptime); 
  	PrintOut(0, buf);
    sprintf(buf, "Filter       : %s", filter); 
  	PrintOut(0, buf);
  	sprintf(buf, "Image shift  : %d, %d pixels", offsetx, offsety); 
  	PrintOut(0, buf);

    if (nx<=0 || ny<=0 || nx>=8192 || ny>=8192) {
	    /* check image size */
	    ffclos(fits,&istat);
	    return MPK_INVALID_SIZE;
    }
    
    /* Read image data */
    ccd = malloc(nx*ny*sizeof(double));
    ffg2dd(fits,1,0.0,nx,nx,ny,ccd,&anyf,&istat);
    if (istat!=0) {
	    ffclos(fits,&istat);
	    return istat;
    }

    if (kombine_nx==0 && kombine_ny==0) {
	    /* first image, sets the variables */
	    kombine_nx = nx;
	    kombine_ny = ny;
	    kombine_jd = cmpack_strtojd(datestr, timestr) + (0.5*exptime/86400.0);
	    kombine_exptime = exptime;
	  	kombine_nframe = 1;
	  	
	  	/* Copy header to destination file */
		ffcopy(fits,kombine_outf,1,&kombine_ostat);	  	
	  
	  if (kombine_bitpix==0) {
		  /* Get equivalent data type and determine the max. and min. value */ 
	  	ffgiet(fits,&kombine_out_bitpix,&istat);
  	  } else {
	    kombine_out_bitpix = kombine_bitpix;
      } 	  
	  switch (kombine_out_bitpix) 
	  {
	  case BYTE_IMG:  
	    kombine_minvalue = 0.0;
	    kombine_maxvalue = 255.0;
	    break;
	  case SBYTE_IMG:  
	    kombine_minvalue = -128.0;
	    kombine_maxvalue = 127.0;
	    break;
	  case SHORT_IMG:
	    kombine_minvalue = -32768.0;
	    kombine_maxvalue = 32767.0;
	    break;
	  case USHORT_IMG:
	    kombine_minvalue = 0.0;
	    kombine_maxvalue = 65535.0;
	    break;
	  case LONG_IMG:
	    kombine_minvalue = -2147483648.0;
	    kombine_maxvalue = 2147483647.0;
	    break;
	  case ULONG_IMG:
	    kombine_minvalue = 0.0;
	    kombine_maxvalue = 4294967296.0;
	    break;
	  default:
	    kombine_minvalue = 0.0;
	    kombine_maxvalue = 65535.0;
	  }
		sprintf(buf, "Data format  : %d", kombine_out_bitpix); 
		PrintOut(1, buf);
		sprintf(buf, "Min. value   : %.1f", kombine_minvalue); 
		PrintOut(1, buf);
		sprintf(buf, "Max. value   : %.1f", kombine_maxvalue); 
		PrintOut(1, buf);
				
	 	/* Init accumulation buffer */
	 	kombine_data = malloc(nx*ny*sizeof(double));
	 	kombine_stat = malloc(nx*ny*sizeof(char));
	 	if (!kombine_data || !kombine_stat) {
		    ffclos(fits, &istat);
		    return MPK_MEMORY;
	 	}
	  	for (j=0;j<ny;j++) {
	      	for (i=0;i<nx;i++) {
	          	ii = i + offsetx;
		        jj = j + offsety;
		        if (ii<0 || jj<0 || ii>=nx || jj>=ny) {
		      		S(i,j) = -2;
		      		D(i,j) = 0;
	      	    } else {
	               	val = I(ii,jj); 
	              	if (val>=kombine_maxvalue) {
		              	D(i,j) = 0;
	              		S(i,j) = 1;
	           		} else 
	           		if (val<=kombine_minvalue) {
		               	D(i,j) = 0;
		            	S(i,j) = -1;
	            	} else {
		            	S(i,j) = 0;
		            	D(i,j) = val;
	            	}
	  	      	}
	      	}
      	}
  	} else {
	  	if (nx!=kombine_nx || ny!=kombine_ny) {
		    /* All input images must have the same size */
		    ffclos(fits,&istat);
		    return KMB_DIFFERENET_SIZE;
	  	}
	  	
	  	/* next image, add data to the accumulation buffer */
    	kombine_jd += cmpack_strtojd(datestr, timestr) + (0.5*exptime/86400.0);
    	kombine_exptime += exptime;
  		kombine_nframe++;
  		
	 	/* Add to the accumulation buffer */
	  	for (j=0;j<ny;j++) {
	      	for (i=0;i<nx;i++) {
		        if (S(i,j)==0) {
	          		ii = i + offsetx;
		        	jj = j + offsety;
			        if (ii<0 || jj<0 || ii>=nx || jj>=ny) {
			      		S(i,j) = -2;
		      	    } else {
			      	    val = I(ii,jj);
		              	if (val>=kombine_maxvalue) {
			            	S(i,j) = 1;
		               	} else
		              	if (val<=kombine_minvalue) {
			            	S(i,j) = -1;
		            	} else {
			            	D(i,j) += val;
		            	}
	            	}
	  	      	}
	      	}
	  	}
  	}
  	free(ccd);
	 	
    /* Normal return */
    ffclos(fits,&istat);
    return 0;
}

int kombine_close()
// Computes the output image and writes it to the file
{
  	int i, j, nx, ny, res = 0;
  	int underflow = 0, overflow = 0, overexposed = 0, badpixels = 0;
  	char buf[MAXLINE], datestr[MAXLINE], timestr[MAXLINE];
  	time_t tsec;
  	struct tm *actim;
  	double val, exptime, jd;

  	/* Check data */
  	if (!kombine_outf || !kombine_fname) {
	  	res = KMB_OUT_NOT_OPEN;
	  	goto close_finish;
  	}
  	if (kombine_nframe<=0 || !kombine_data || !kombine_stat) {
	  	res = MPK_NO_INPUT_FILES;
	  	goto close_finish;
	}

	/* Compute header information */
    exptime = kombine_exptime / kombine_nframe;
    jd      = kombine_jd/kombine_nframe - (0.5*exptime/86400.0);
  	cmpack_jdtostr(jd, datestr, timestr);
	
  	PrintOut(0, "----------------------------------------------");
  	sprintf(buf,"Output file  : %s", kombine_fname);
  	PrintOut(0, buf);
  	sprintf(buf,"No. of frames: %d frames" , kombine_nframe); 
  	PrintOut(0, buf);
  	sprintf(buf,"Date & time  : %s %s UT" , datestr, timestr); 
  	PrintOut(0, buf);
  	sprintf(buf,"Exposure     : %.2f s", exptime); 
  	PrintOut(0, buf);
  
  	/* Write header information. Delete other known keys, which carry date 
  	   and time of observation, because they would hold wrong value */
  	kombine_ostat=0; ffukyj(kombine_outf,"FR_AVG",kombine_nframe,"Number of frames averaged",&kombine_ostat);
  	kombine_ostat=0; ffukyg(kombine_outf,"EXPTIME",exptime,2, "EXPOSURE IN SECONDS",&kombine_ostat);
  	kombine_ostat=0; ffukys(kombine_outf,"DATE-OBS",datestr,"UT DATE OF START",&kombine_ostat);
  	kombine_ostat=0; ffukys(kombine_outf,"TIME-OBS",timestr,"UT TIME OF START",&kombine_ostat);
  	kombine_ostat=0; ffdkey(kombine_outf,"D02E-OBS",&kombine_ostat); 
  	kombine_ostat=0; ffdkey(kombine_outf,"DATE-BEG",&kombine_ostat); 
  	kombine_ostat=0; ffdkey(kombine_outf,"TM-START",&kombine_ostat); 
  	kombine_ostat=0; ffdkey(kombine_outf,"TM-MID",&kombine_ostat); 
  	kombine_ostat=0; ffdkey(kombine_outf,"TM-END",&kombine_ostat); 
  	kombine_ostat=0; ffdkey(kombine_outf,"REC-STRT",&kombine_ostat); 
  	kombine_ostat=0; ffdkey(kombine_outf,"REC-END",&kombine_ostat); 
  	kombine_ostat=0; ffdkey(kombine_outf,"COR-MEAN",&kombine_ostat); 
  	kombine_ostat=0; ffdkey(kombine_outf,"DATE-OB1",&kombine_ostat); 
  	kombine_ostat=0; ffdkey(kombine_outf,"OBS-DATE",&kombine_ostat); 
  	kombine_ostat=0; ffdkey(kombine_outf,"OBS-DAT1",&kombine_ostat); 
 	kombine_ostat=0; ffphis(kombine_outf,"Combined frame.",&kombine_ostat); 
 	kombine_ostat=0;

  	/* Actual current date and origin */
  	tsec = time(&tsec);
  	actim = localtime(&tsec);
  	cmpack_datetostr(actim->tm_year+1900, actim->tm_mon+1, actim->tm_mday, buf);
  	ffukys(kombine_outf,"DATE", buf, "DATE OF CREATION",&kombine_ostat);
  	sprintf(buf,"%s (%s) %s", PROGNAME, PKGNAME, VERSION);
  	ffukys(kombine_outf,"ORIGIN",buf,"",&kombine_ostat);
 	kombine_ostat=0;

  	/* Computation output image data */
  	nx = kombine_nx;
  	ny = kombine_ny;
  	for (j=0;j<ny;j++) {
      	for (i=0;i<nx;i++) {
          	/* Compute the output value */
          	val = D(i,j)/kombine_nframe;
          	if (S(i,j)==-2) {
	          	D(i,j) = 0.0;
          	} else
          	if (S(i,j)==-1) {
	          	D(i,j) = kombine_minvalue;
	          	badpixels = 1;
          	} else
          	if (S(i,j)==1) {
	          	D(i,j) = kombine_maxvalue;
      			overexposed = 1;
          	} else
          	if (val<kombine_minvalue) {
	          	D(i,j) = kombine_minvalue;
	          	underflow = 1;
          	} else
          	if (val>kombine_maxvalue) {
	          	D(i,j) = kombine_maxvalue;
	          	overflow = 1;
          	} else {
	          	D(i,j) = val;
          	}
      	}
  	}
  	/* Write image data */
  	ffp2dd(kombine_outf,1,nx,nx,ny,kombine_data,&kombine_ostat);

  	if (overexposed) {
	  	sprintf(buf,"There are overexposed pixels in the source frames (max.=%.1f).", kombine_minvalue);
	  	PrintOut(1, buf);
  	}
  	if (badpixels) {
   	  	sprintf(buf,"There are bad pixels in the source frames (min.=%.1f).", kombine_maxvalue);
	  	PrintOut(1, buf);
    }
  	if (overflow) {
	  	sprintf(buf,"An overflow has been occurred during computation (max.=%.1f).", kombine_maxvalue);
	  	PrintOut(1, buf);
  	}
  	if (underflow) {
   	  	sprintf(buf,"An underflow has been occurred during computation (min.=%.1f).", kombine_minvalue);
	  	PrintOut(1, buf);
  	}
  	
  	/* Normal return */
close_finish:
  	if (kombine_data) {
  		free(kombine_data);
  		kombine_data = NULL;
	}
	if (kombine_stat) {
  		free(kombine_stat);
  		kombine_stat = NULL;
	}
	if (kombine_outf) {
  		ffclos(kombine_outf, &kombine_ostat);
  		kombine_outf = NULL;
	}
	if (res!=0 && kombine_fname) {
		remove(kombine_fname);
	}
	if (kombine_fname) {
  		free(kombine_fname);
  		kombine_fname = NULL;
	}
  	return res;
}

int kombine_clean(void)
// Frees allocated memory
{
	if (kombine_outf) {
		ffclos(kombine_outf, &kombine_ostat); 
		kombine_outf = NULL;
	}
	if (kombine_data) {
		free(kombine_data); 
		kombine_data = NULL;
	}
	if (kombine_stat) {
		free(kombine_stat);
		kombine_stat = NULL;
	}
	if (kombine_fname) {
		free(kombine_fname);
		kombine_fname = NULL;
	}
	kombine_nframe 	= 0;
  	kombine_nx      = 0;
  	kombine_ny      = 0;
  	kombine_ostat 	= 0;
  	kombine_jd      = 0.0;
  	kombine_exptime = 0.0;
  	kombine_nframe 	= 0;
  	return 0;
}
