/**************************************************************

darkbat.c (C-Munipack project)
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <fitsio.h>
#include <math.h>
#include <time.h>

#include "comfun.h"
#include "fitsfun.h"
#include "imglist.h"
#include "robmean.h"
#include "inifile.h"
#include "mdark.h"

int   mdark_nx;                // Width of frames
int   mdark_ny;                // Height of frames
double mdark_minvalue;          // Lowest good datum
double mdark_maxvalue;          // Highest good datum
double mdark_exp;               // Exposure duration
double mdark_temp;              // Temperature
char  mdark_date[MAXLINE];     // Date observation of first frame
char  mdark_time[MAXLINE];     // Time observation of first frame
tImgList mdark_list = NULL;    // List of source frames
int   mdark_bitpix;            // Output image data format
int   mdark_mbias;			   // Make bias frame instead of dark frame
int   mdark_scalable;		   // Make scalable dark
double *mdark_bdata;           // Data of bias-frame
int   mdark_biasnx;			   // Width of bias frame
int   mdark_biasny;            // Height of bias frame
int   mdark_out_bitpix;			// Output data format

static int mdark_initbias(const char *bias)
// Read bias frame
{
	FILE *f;
    fitsfile *dfits = NULL;
    long naxes[2];
 	char buf[MAXLINE];
 	int  dstat = 0;
    int anyf, naxis, bitpix, simple, extend;
    long pcount, gcount;

    /* Check parameters */
    if (!bias) {
        PrintOut(0, "Missing name of the bias frame.");
 	    return MPK_INVALID_PAR;
    }
    if (strlen(bias)==0) {
        PrintOut(0, "Missing name of the bias frame.");
 	    return MPK_INVALID_PAR;
    }
    f = fopen(bias,"r");
    if (f==NULL) {
 	    PrintOut(1, "Bias frame file not found.");
 	    return MDK_CANNOT_OPEN_BIAS;
    }
    fclose(f);

    /* Open bias file */
    sprintf(buf,"Using bias frame: %s",bias);
    PrintOut(0, buf);
    
    ffdkopn(&dfits,bias,READONLY,&dstat);
    if (dstat!=0) return dstat;

    /* Read and check dimensions */
	ffghpr(dfits,2,&simple,&bitpix,&naxis,naxes,&pcount,&gcount,&extend,&dstat);
	if (dstat!=0) return dstat;
	mdark_biasnx = naxes[0];
	mdark_biasny = naxes[1];
	if ((mdark_biasnx<=0)||(mdark_biasny<=0)) {
		ffclos(dfits,&dstat);
		return MPK_INVALID_SIZE;
	}

	/* Allocate the memory for bias frame */
	mdark_bdata = malloc(mdark_biasnx*mdark_biasny*sizeof(double));
	if (!mdark_bdata) {
		ffclos(dfits,&dstat);
		return MPK_MEMORY;
	}

	/* Read bias frame data */
    ffg2dd(dfits,1,0.0,mdark_biasnx,mdark_biasnx,mdark_biasny,mdark_bdata,&anyf,&dstat);
	if (dstat!=0) return dstat;

    /* Close bias file */
    ffclos(dfits,&dstat);
    if (dstat!=0) return dstat;
    
    return 0;
}	

int mdark_init(const char *optfile, int flags, const char *biasfile)
// Initializes the parameters and the image list
{
  int res;
  inifile *f;
  char msg[80+FILENAME_MAX];

  mdark_nx      = 0;
  mdark_ny      = 0;
  mdark_list    = ImgList_Create(0, 0);
  mdark_exp     = 0.0;
  mdark_temp    = 0.0;
  mdark_date[0] = 0;
  mdark_time[0] = 0;
  mdark_bdata   = NULL;
  mdark_mbias   = (flags & MPK_MDARK_MASTERBIAS)!=0;
  mdark_scalable = !mdark_mbias && (flags & MPK_MDARK_SCALABLE)!=0;
  mdark_biasnx  = 0;
  mdark_biasny  = 0;
  
  /* Read parameter file */
  f = iniopen(optfile, OPTFILE);
  if (f==NULL) {
	if (optfile) sprintf(msg,"Cannot read parameter file '%s'", optfile);
		    else sprintf(msg,"Cannot read parameter file '%s'", OPTFILE);
	PrintOut(0, msg);
  	return MPK_CANNOT_OPEN_PAR;
  }
  mdark_bitpix = inigkyl(f, "BITPIX", 0);
  iniclos(f);
  
  /* Print parameters */
  PrintOut(1, "Configuration parameters:");
  sprintf(msg, "BITPIX: %d", mdark_bitpix);
  PrintOut(1, "-------------------------------");
    
  /* Read bias frame (if we are making scalable dark) */
  if (mdark_scalable) {
	  res = mdark_initbias(biasfile);
	  if (res!=0) return res;
  }

  return 0;
}

int mdark_read(const char *filename)
// Reads the input frame and stores it in the image list
{
	FILE *f;
	int anyf, nx, ny, bitpix, istat = 0;
	fitsfile *fits;
    double *ccd;
    double exptime, temp;
    char buf[MAXLINE], date[MAXLINE], time[MAXLINE];

    /* Check source file */
    sprintf(buf,"Image file   : %s",filename);
    PrintOut(0, buf);
    if (!filename) {
        PrintOut(0, "Missing name of the source image file.");
 	    return MPK_INVALID_PAR;
    }
    if (strlen(filename)==0) {
        PrintOut(0, "Missing name of the source image file.");
 	    return MPK_INVALID_PAR;
    }
    f = fopen(filename,"r");
    if (f==NULL) {
        PrintOut(1, "Source image file not found.");
        return MPK_CANNOT_OPEN_SRC;
    }
    fclose(f);

    /* open input file */
    ffdkopn(&fits,filename,READWRITE,&istat);
    if (istat!=0) {
	    ffclos(fits, &istat);
	    return istat;
    }

    /* Read parameters of source file */
    fits_gpar(fits, &nx, &ny, &bitpix);
	fits_gexp(fits, &exptime);
	fits_gtemp(fits, &temp);
    fits_getdatetime(fits, date, time);

    if ((nx==0)||(ny==0)) {
	    ffclos(fits,&istat);
	    return MPK_INVALID_SIZE;
    }
    if ((mdark_nx==0)&&(mdark_ny==0)) {
        /* first image */
        mdark_nx = nx;
        mdark_ny = ny;
        mdark_exp  = exptime;
        mdark_temp = temp;
        strcpy(mdark_date, date);
        strcpy(mdark_time, time);
        if (mdark_bitpix==0) {
	        /* The output frame have the same format as the first source frame */
	        ffgiet(fits, &mdark_out_bitpix, &istat);
        } else {
	        /* Use format from the configuration file */
	        mdark_out_bitpix = mdark_bitpix;
        }
		switch (mdark_out_bitpix) 
		{
		case BYTE_IMG:  
			mdark_minvalue = 0.0;
			mdark_maxvalue = 255.0;
			break;
		case SBYTE_IMG:  
			mdark_minvalue = -128.0;
			mdark_maxvalue = 127.0;
			break;
		case SHORT_IMG:
			mdark_minvalue = -32768.0;
			mdark_maxvalue = 32767.0;
			break;
		case USHORT_IMG:
			mdark_minvalue = 0.0;
			mdark_maxvalue = 65535.0;
			break;
		case LONG_IMG:
			mdark_minvalue = -2147483648.0;
			mdark_maxvalue = 2147483647.0;
			break;
		case ULONG_IMG:
			mdark_minvalue = 0.0;
			mdark_maxvalue = 4294967296.0;
			break;
		default:
			mdark_minvalue = 0.0;
			mdark_maxvalue = 65535.0;
		}
		sprintf(buf, "Data format  : %d", mdark_out_bitpix); 
		PrintOut(1, buf);
		sprintf(buf, "Min. value   : %.1f", mdark_minvalue); 
		PrintOut(1, buf);
		sprintf(buf, "Max. value   : %.1f", mdark_maxvalue); 
		PrintOut(1, buf);
    }
    if (mdark_scalable && (nx!=mdark_biasnx || ny!=mdark_biasny)) {
	    ffclos(fits,&istat);
	    return MDK_DIFF_BIASSIZE;
    }
    if ((nx!=mdark_nx)||(ny!=mdark_ny)) {
	    ffclos(fits,&istat);
	    return MDK_DIFFERENET_SIZE;
    }

    /* image data input */
    ccd = malloc(nx*ny*sizeof(double));
    ffg2dd(fits,1,0.0,nx,nx,ny,ccd,&anyf,&istat);
    if (istat!=0) {
	    ffclos(fits,&istat);
	    return istat;
    }
    ffclos(fits,&istat);

    /* Add this image to image list */
    if (!ImgList_Add(mdark_list, ccd)) {
        free(ccd);
        return MPK_MEMORY;
    }

    /* Normal return */
    return 0;
}

int mdark_write(const char *filename)
{
  fitsfile *fits;
  int istat = 0, n, i, j, k;
  double a, b, *fbuf, *dark, **ccds;
  long naxes[2], nx, ny, underflow = 0, overflow = 0;
  char buf[MAXLINE], actdate[MAXLINE];
  time_t tsec;
  struct tm *actim;

  n  = ImgList_Count(mdark_list);
  if (n<=0) return MPK_NO_INPUT_FILES;

  /* makes array of images for faster access to data */
  ccds = malloc(n*sizeof(double*));
  ImgList_First(mdark_list);
  for (k=0; k<n; k++) {
	  ccds[k] = ImgList_GetImage(mdark_list);
      ImgList_Next(mdark_list);
  }

  /* Computation of mean of values for each pixel */
  nx = mdark_nx;
  ny = mdark_ny;
  dark = malloc(nx*ny*sizeof(double));
  fbuf = malloc(n*sizeof(double));
  for (j=0;j<ny;j++) {
      for (i=0;i<nx;i++) {
	      // Computation of mean
          for (k=0;k<n;k++) {
	          fbuf[k]=ccds[k][i+j*nx];
          }
          // Compute their median 
          cmpack_robustmeanD(n, fbuf, &a, &b);
          // Subtract bias 
          if (mdark_scalable) {
	          a -= mdark_bdata[i+j*nx];
          }
          // Range checking
          if (a<mdark_minvalue) {
	           underflow++;
               a = mdark_minvalue;
          } else
          if (a>mdark_maxvalue) {
               overflow++;
               a = mdark_maxvalue;
          }
          // Store the mean into output frame
          dark[i+j*nx] = a;
      }
  }
  free(fbuf);
  free(ccds);

  /* Open output file */
  PrintOut(0, "-------------------------------");
  sprintf(buf,"Output file  : %s",filename);
  PrintOut(0, buf);
  remove(filename);
  ffdkinit(&fits,filename,&istat);

  /* Actual date */
  tsec = time(&tsec);
  actim = localtime(&tsec);
  sprintf(actdate,"%04d-%02d-%02d",actim->tm_year+1900,actim->tm_mon+1,actim->tm_mday);

  /* Write header information */
  naxes[0]=nx; naxes[1]=ny;
  ffphps(fits,mdark_out_bitpix,2,naxes,&istat);
  if (mdark_mbias)
 	  ffpkys(fits,"OBJECT","Bias-frame"," ",&istat);
  else
      ffpkys(fits,"OBJECT",(mdark_scalable?"Scalable dark-frame":"Dark-frame")," ",&istat);
  if (mdark_exp>=0.0)
    ffpkyg(fits,"EXPTIME",mdark_exp,2,"EXPOSURE IN SECONDS",&istat);
  if ((mdark_temp>=-999.0)&&(mdark_temp<=999.0))
    ffpkyg(fits,"TEMPERAT",mdark_temp,2,"CCD TEMPERATURE",&istat);
  if (mdark_scalable) 
    ffphis(fits,"Bias-frame subtracted.",&istat);
  ffpkys(fits,"DATE",actdate,"DATE OF CREATION",&istat);
  sprintf(buf,"%s (%s) %s", (mdark_mbias ? MBIAS_PROGNAME : MDARK_PROGNAME), PKGNAME, VERSION);
  ffpkys(fits,"ORIGIN",buf,"",&istat);
  ffpkys(fits,"DATE-OBS",mdark_date,"UT DATE OF FIRST FRAME (ASSUMES PC CLOCK=UT)",&istat);
  ffpkys(fits,"TIME-OBS",mdark_time,"UT TIME OF FIRST FRAME (ASSUMES PC CLOCK=UT)",&istat);
  if (!mdark_mbias) {
	  if (mdark_scalable)
      	ffpkyl(fits,"SCALABLE",mdark_scalable,"BIAS FRAME WAS SUBTRACTED",&istat);
      else
      	ffpkyl(fits,"SCALABLE",mdark_scalable,"BIAS FRAME WASN'T SUBTRACTED",&istat);
  }

  /* Write image data */
  ffp2dd(fits,1,nx,nx,ny,dark,&istat);
  if (istat!=0) return istat;

  /* Close the output file */
  ffclos(fits,&istat);
  if (istat!=0) return istat;

  /* Range check */
  if (overflow>0) {
     sprintf(buf,"An overflow has been occurred on %ld of %ld pixels during computation (max.=%.1f).", 
     	overflow, nx*ny, mdark_maxvalue);
     PrintOut(0, buf);
  }
  if (underflow>0) {
  	 sprintf(buf,"An underflow has been occurred on %ld of %ld pixels during computation (min.=%.1f).", 
  	 	underflow, nx*ny, mdark_minvalue);
     PrintOut(0, buf);
  }
  
  /* Return value */
  return 0;
}

int mdark_clean(void)
// Frees allocated images and image list
{
	double *ccd;
	
	/* Free allocated images and clear the image list */
    if (ImgList_First(mdark_list)) {
        do {
	        ccd = ImgList_GetImage(mdark_list);
            if (ccd) free(ccd);
            ImgList_SetImage(mdark_list, NULL);
        } while (ImgList_Next(mdark_list));
    }
	ImgList_Destroy(mdark_list);
	mdark_list = NULL;
	
	/* Free bias frame */
	if (mdark_bdata) {
		free(mdark_bdata);
		mdark_bdata = NULL; 
		mdark_biasnx = mdark_biasny = 0;
	}
	
	/* Normal return */
	return 0;
}
