/**************************************************************

varfind.c (C-Munipack project)
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <math.h>

#ifdef _WIN32
#include <io.h>
#endif

#include "comfun.h"
#include "trajd.h"
#include "mfind.h"

void PrintHelp()
{
 printf("MUNIFIND version %s - computing table for finding variables.\n",VERSION);
 printf(PROJECTLINE);
 printf("\nUsage: munifind [options] [@dirfile] [ref=n] [thr=x] [col=n] [image1 ...]\n\n");
 printf("parameters:\n");
 printf("     @dirfile    file with input image names\n");
 printf("                 (for only '@', read filenames from standart input)\n");
 printf("     ref=n       index of reference star (default = autodetect)\n");
 printf("     par=file    parameter file (default is %s)\n",OPTFILE);
 printf("     src=file    all-stars file\n");
 printf("     thr=x       threshold in %%\n");
 printf("     col=n       column for listing\n");
 printf("     ap=n        aperture number\n");
 printf("     out=file    output file (default = %s)\n",OUTFILE);
 printf("     image1...   list of input photometry files\n");
 printf("options:\n");
 printf("     -p<file>    parameter file (equivalent with 'par=')\n");
 printf("     -a<file>    all-stars file (equivalent with 'src=')\n");
 printf("     -q          quiet mode\n");
 printf("     -h          give this help\n");
 printf("     -v          prints software version\n");
 printf("     -l          display software license\n");
 printf("     -g          creates the example of the parameter file\n");
 printf(WARRANTYLINE);
 printf(HOMEPAGELINE);
 printf(BUGREPORTS);
}

void report(int status, const char *file)
{
 if (cmpack_debug)
 	fprintf(stderr, "The following error(s) has been occurred: %d\n",status);
 PrintError(PROGNAME, file, 0, status);
}

void error(int status, const char *file)
{
 report(status, file);
 exit(status);
}

void CreatePar()
/* Creates a sample parameter file */
{
 FILE *fp;

 if ((fp=fopen(SMPFILE,"w"))!=NULL) {
   printf("Writing '%s' ...\n\n",SMPFILE);
   fprintf(fp,"APERTURE = %d     # Aperture number\n",DEFAPERTURE);
   fprintf(fp,"THRESHOLD = %d    # Threshold in %%\n",DEFTHRESHOLD);
   fclose(fp);
   printf("'%s' is an example of the parameter file. Rename it to '%s' and edit the parameters.\n\n",SMPFILE,OPTFILE);
 }
}

int main(int argc, char **argv)
{
    int i, n, res, stdinput = 0, ap = -1, ref = -1, thr = -1, level = 0;
    char line[MAXLINE], buf1[MAXLINE], *par;
	char *namsci = "", *namsrc = "", *namout = OUTFILE, *namopt = NULL;
	FILE *f;

	/* Zpracovani parametru prikazove radky */
	for (i=1;i<argc;i++) {
	   par = argv[i];
	   /* Prints help */
	   if ((strcmp(par,"-h")==0)||(strcmp(par,"--help")==0)) {
		  PrintHelp();
          return 0;
       }
	   /* Prints version */
	   if ((strcmp(par,"-v")==0)||(strcmp(par,"--version")==0)) {
		  PrintVersion(PROGNAME);
          return 0;
       }
       /* Prints license */
       if ((strcmp(par,"-l")==0)||(strcmp(par,"--license")==0)) {
          PrintLicense();
          return 0;
       }
       /* Generate sample parameter file */
       if (strcmp(par,"-g")==0) {
           CreatePar();
           return 0;
       }
       /* Flag for input from stdin or dirfile */
       if (strchr(par,'@')==par) {
	       namsci = malloc(strlen(par));
	       strcpy(namsci,&par[1]);
           stdinput = strlen(namsci)==0;
       }
       /* Output file */
       if (strstr(par,"out=")==par) {
	       namout = malloc(strlen(par)-3);
	       strcpy(namout,&par[4]);
       }
       /* Aperture selection */
       if (strstr(par,"col=")==par) {
	       ap = (atol(par+4)-2)/2;
	       if (ap<=0) ap = 1;
       }
       if (strstr(par,"ap=")==par) {
	       ap = atol(par+3);
	       if (ap<=0) ap = 1;
       }
       /* Threshold */
       if (strstr(par,"thr=")==par) {
	       thr = atol(par+4);
	       if (thr<=0) thr = 60;
       }
       /* Reference star selection */
       if (strstr(par,"ref=")==par) {
	       ref = atol(par+4);
	       if (ref<=0) ref = -1;
       }
       /* Parameter file */
       if (strstr(par,"par=")==par) {
	       namopt = malloc(strlen(par)-3);
	       strcpy(namopt,&par[4]);
       }
       if (strstr(par,"-p")==par) {
	       namopt = malloc(strlen(par)-1);
	       strcpy(namopt,&par[2]);
       }
       /* Source all-stars file */
       if (strstr(par,"all=")==par) {
	       namsrc = malloc(strlen(par)-3);
	       strcpy(namsrc,&par[4]);
       }
       if (strstr(par,"-a")==par) {
	       namsrc = malloc(strlen(par)-1);
	       strcpy(namsrc,&par[2]);
       }
       /* Quiet/verbose flags */
       if (strcmp(par,"-q")==0 || strcmp(par,"--quiet")==0 || strcmp(par,"--silent")==0) {
           level = MPK_LEVEL_QUIET;
       }
       if (strcmp(par,"--verbose")==0 || strcmp(par,"--debug")==0) {
           level = MPK_LEVEL_DEBUG;
       }
  }

  cmpack_setoutput(PrintMessage, level);
  PrintPrologue(PROGNAME);
  res = mfind_init(namopt, ap, thr);
  if (res!=0) error(res, NULL);
  n = 0;

  /* image names reads from command line */
  for (i=1;i<argc;i++) {
	  par = argv[i];
      if ((strchr(par,'=')==0)&&(strchr(par,'@')!=par)&&(strchr(par,'-')!=par)&&(atol(par)==0)) {
#ifdef _WIN32
	  intptr_t fh;
    struct _finddata_t fd;
		 if ((fh=_findfirst(par, &fd))!=-1) {
			do {
				if ((fd.attrib & _A_SUBDIR)==0)  {
          			res=mfind_readsrt(fd.name);
          			if (res!=0) report(res, fd.name);
          			n++;
          		}
          	} while (_findnext(fh, &fd)==0);
		 	_findclose(fh);
      	 } else {
	      	 report(MPK_CANNOT_OPEN_SRC, par);
      	 }
#else
		res=mfind_readsrt(par);
		if (res!=0) report(res, par);
		n++;
#endif
      }
  }

  /* image names reads from dirfile */
  if (strlen(namsci)>0) {
	  f = fopen(namsci,"r");
	  if (f==NULL) error(MPK_NO_DIRFILE, namsci);
      while(fgets(buf1,MAXLINE,f)!=NULL) {
          sscanf(buf1,"%s",line);
          if (strcmp(line,"")) {
	          res=mfind_readsrt(line);
    	      if (res) report(res, line);
	          n++;
          }
      }
  }

  /* image names reads from all-stars file */
  if (strlen(namsrc)>0) {
      res=mfind_readall(namsrc);
      if (res) report(res, namsrc);
      n++;
  }

  /* Image names reads from stdin */
  if (stdinput) {
      while(fgets(buf1,MAXLINE,stdin)!=NULL) {
          sscanf(buf1,"%s",line);
          if (strcmp(line,"")) {
	          res=mfind_readsrt(line);
    	      if (res) report(res, line);
	          n++;
          }
      }
  }

  if (n<=0) {
	  PrintOut(0, "No input files.");
	  res = MPK_NO_DATA;
  }
  if (res==0) {
	  res = mfind_compute(ref);
	  if (res!=0) report(res, NULL);
  }
  if (res==0) {
      res = mfind_write(namout);
      if (res!=0) report(res, namout);
  }

  /* Normal return */
  res=mfind_clean();
  if (res!=0) error(res, NULL);

  return 0;
}
