/**************************************************************

mlist.c (C-Munipack project)
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "comfun.h"
#include "convert.h"
#include "trajd.h"
#include "inifile.h"
#include "mlist.h"
#include "mparams.h"

#define	VALID_MAG 0x01
#define VALID_ERR 0x02
#define VALID_RA  0x04
#define VALID_DEC 0x08
#define VALID_LON 0x10
#define VALID_LAT 0x20

struct ColInfo {
	char valname[80];       // Value field name
	char errname[80];       // Error field name
};

struct ColData {
	int ok;
	double val, err;		// Value and error
	struct ColData *next;	// Pointer to next col
};

struct RowData {
	int frame;       		// Index of frame, valid flag
	int ncols;              // Number of columns
	double jd, hcor, amass; // Geocentric julian date and hel. cor.
	double *val, *err;      // Values and standard errors
	int *ok;				// Column validity flags
	double *sum, *sum2;		// Accumulators (for stddev format)
	int *count;				// Number of measurements (for stddev format)
	struct RowData *next;   // Pointer to next row
};

struct StarData {
	int index;
	double xmag[MAXAP], xerr[MAXAP];
	int xvalid[MAXAP];		
	struct StarData *next;	
};

int     ls_aper;        	// Aperture number
char    ls_filter[MAXLINE]; // Filter
int     ls_format;          // Output format
int     ls_jdhel;           // Print heliocentric JD
int     ls_helcor;          // Print heliocentric correction
int     ls_amass;           // Print airmass coefecient
int     ls_maxst;           // Highest index of selected stars
double  ls_ra;              // Object's right ascension
double  ls_dec;             // Object's declination
double  ls_la;              // Object's right ascension
double  ls_be;              // Object's declination
double  ls_lat;             // Observer's latitude
double  ls_lon;             // Observer's longitude
int     ls_ncols;           // Number of fields
int		ls_nstar;			// Number of selected stars
int     ls_objset;			// Validation flag for object's coordinates
int		ls_obsset;			// Validation flag for observer's coordinates
struct  ColInfo *ls_col;    // Field definitions
struct  RowData *ls_first;  // First data row
struct  RowData *ls_last;   // Last data row
int     ls_nvar;            // No. of variable stars
struct  StarData *ls_var, *ls_var_last;   // List of variable stars
int     ls_ncmp;            // No. of comparison stars
struct  StarData *ls_cmp, *ls_cmp_last;   // List of comparison stars
int     ls_nchk;            // No. of check stars
struct  StarData *ls_chk, *ls_chk_last;   // List of check stars
double  *ls_mag;            // Magnitudes in input file
double  *ls_err;            // Errors in input file

void starid(char *buf, int id)
// String reprezentation of the star
{
	if (ls_format==MPK_FORMAT_DIFFMAG) {
		if (id==0) strcpy(buf,"V"); else
		if (id==1) strcpy(buf,"C"); else
		sprintf(buf,"C%d",id-1);
	} else
	if (ls_format==MPK_FORMAT_STDDEV) {
		if (id==0) strcpy(buf,"C"); 
		else sprintf(buf,"C%d",id);
	} else {
		sprintf(buf,"S%d",id+1);
	}
}

static int mlist_cbproc(const char *id, const char *val)
// Parser callback procedure
{
	int i;
	struct StarData *st;
	
	if (strcmp(id, "ra")==0) {
		if (cmpack_strtora(val, &ls_ra)!=0) {
	    	PrintOut(0, "Invalid format of object's coordinates (R.A.)");
      		return MPK_INVALID_PAR;
  	  	} else {
	  	  	ls_objset |= VALID_RA;
  	  	}
	} else
	if (strcmp(id, "dec")==0) {
  	  	if (cmpack_strtodec(val, &ls_dec)!=0) {
	    	PrintOut(0, "Invalid format of object's coordinates (DEC.)");
      		return MPK_INVALID_PAR;
  	  	} else {
	  	  	ls_objset |= VALID_DEC;
  	  	}
  	} else 
  	if (strcmp(id, "lon")==0) {
	  	if (cmpack_strtolon(val, &ls_lon)!=0) {
	    	PrintOut(0, "Invalid format of observer's coordinates (LON.)");
      		return MPK_INVALID_PAR;
  	  	} else {
	  	  	ls_obsset |= VALID_LON;
  	  	}
	} else
	if (strcmp(id, "lat")==0) {
  	  	if (cmpack_strtolat(val, &ls_lat)!=0) {
	    	PrintOut(0, "Invalid format of observer's coordinates (LAT.)");
      		return MPK_INVALID_PAR;
  	  	} else {
	  	  	ls_obsset |= VALID_LAT;
  	  	}
	} else 
	if (strcmp(id, "var")==0) {
		st = malloc(sizeof(struct StarData));
		i = atoi(val);
		if (st && i>0) {
			st->index = i;
			st->next = NULL;
			if (ls_var_last==NULL) {
				ls_var_last = ls_var = st; 
			} else {
				ls_var_last->next = st; 
				ls_var_last = st;
			}
			ls_nvar++;
		}
	} else 
	if (strcmp(id, "comp")==0) {
		st = malloc(sizeof(struct StarData));
		i = atoi(val);
		if (st && i>0) {
			st->index = i;
			st->next = NULL;
			if (ls_cmp_last==NULL) {
				ls_cmp_last = ls_cmp = st; 
			} else {
				ls_cmp_last->next = st; 
				ls_cmp_last = st;
			}
			ls_ncmp++;
		}
	} else 
	if (strcmp(id, "chk")==0) {
		st = malloc(sizeof(struct StarData));
		i = atoi(val);
		if (st && i>0) {
			st->index = i;
			st->next = NULL;
			if (ls_chk_last==NULL) {
				ls_chk_last = ls_chk = st; 
			} else {
				ls_chk_last->next = st; 
				ls_chk_last = st;
			}
			ls_nchk++;
		}
	}
  	return 0;
}

int mlist_init(const char *optfile, int ap, int outfrm, const char *params)
// Allocates the memory, initializes parameters and opens output file
{
  char *alat, *alon, buf1[MAXLINE], buf2[MAXLINE], msg[MAXLINE];
  int i,j,k,n,res,jdform,helcor,airmass;
  inifile *f = NULL;
  struct StarData *st;

  // Init star list, field list and data list
  ls_ra = ls_dec = ls_lat = ls_lon = 0.0;
  ls_maxst = ls_ncols = 0;
  ls_mag = ls_err = NULL; 
  ls_col = NULL; 
  ls_first = ls_last = NULL;
  ls_nvar = ls_ncmp = ls_nchk = 0;
  ls_var = ls_cmp = ls_chk = NULL;
  ls_var_last = ls_cmp_last = ls_chk_last = NULL;
  ls_objset = ls_obsset = 0;
  strcpy(ls_filter, "");
  
  // Read parameter file
  f = iniopen(optfile, OPTFILE);
  if (f==NULL) {
      if (optfile) sprintf(msg,"Cannot read parameter file '%s'", optfile);
	          else sprintf(msg,"Cannot read parameter file '%s'", OPTFILE);
      PrintOut(0, msg);
      return MPK_CANNOT_OPEN_PAR;
  }
  ls_aper   = inigkyl(f, "APERTURE", DEFAPERTURE);
  ls_format = inigkyl(f, "OUTFORM",  DEFOUTFORM);
  jdform    = inigkyl(f, "JDFORM", 0);
  helcor    = inigkyl(f, "HELCOR", 0);
  airmass   = inigkyl(f, "AIRMASS", 0);
  alon      = inigkys(f, "LON", NULL);
  alat      = inigkys(f, "LAT", NULL);
  iniclos(f);
  
  	if (alon) {
		if (cmpack_strtolon(alon, &ls_lon)==0)
			ls_objset |= VALID_LON;
	}
	if (alat) {
		if (cmpack_strtolon(alat, &ls_lat)==0)
			ls_objset |= VALID_LAT;
	}

  // Process additional parameters
  if (params) {
	  res = cmpack_parse_params(params, mlist_cbproc);
	  if (res!=0) {
		  PrintOut(0, "Syntax error in custom parameter string");
		  PrintOut(1, params);
	      return MPK_INVALID_PAR;
		}
  }
  
  // Function parameter overwrites the value in parameter file
  if (ap>0)      
      ls_aper   = ap;
      
  if (outfrm>=0) {
      ls_jdhel  = (outfrm & MPK_FORMAT_JDHEL)!=0;
      ls_helcor = (outfrm & MPK_FORMAT_HELCOR)!=0;
      ls_amass  = (outfrm & MPK_FORMAT_AIRMASS)!=0;
      ls_format = (outfrm & 0xFF);
  } else {
      ls_jdhel  = (jdform==1);
      ls_helcor = (helcor>0);
      ls_amass  = (airmass>0);
      ls_format = (ls_format & 0xFF);
  }

    sprintf(msg, "APERTURE : %d", ls_aper); 
	PrintOut(1, msg);
	sprintf(msg, "OUTFORM  : %s %s %s %s", 
	  (ls_format==MPK_FORMAT_DIFFMAG ? "dif" :
	  (ls_format==MPK_FORMAT_INSTMAG ? "abs" :
	  (ls_format==MPK_FORMAT_READALL ? "readall" :
	  (ls_format==MPK_FORMAT_TRACKLST ? "trklist" :
	  (ls_format==MPK_FORMAT_STDDEV ? "stddev" : "unknown"))))), 
	  (ls_jdhel ? "jdhel" : ""),
	  (ls_helcor ? "helcor" : ""), 
	  (ls_amass ? "amass" : "")
	);
	PrintOut(1, msg);
  
  
  // Check parameters
  if (ls_aper<=0) {
	  sprintf(msg, "Invalid aperture number (%d)",ls_aper);
	  PrintOut(0, msg);
	  return MPK_INVALID_PAR;
  }
  if (ls_jdhel || ls_helcor || ls_amass) {
	if ((ls_objset & VALID_RA)==0 || (ls_objset & VALID_DEC)==0) {
      PrintOut(0, "Object's coordinates must be specified.");
      return MPK_INVALID_PAR;
    } else {
	  cmpack_rdtolb(ls_ra, ls_dec, &ls_la, &ls_be);
  	  cmpack_ratostr(ls_ra, buf1);
	  cmpack_dectostr(ls_dec, buf2);
	  sprintf(msg,"Object's coordinates are R.A.=%s and DEC.=%s.", buf1, buf2);
	  PrintOut(1, msg);
    }
  }
  if (ls_amass) {
	if ((ls_obsset & VALID_LAT)==0 || (ls_obsset & VALID_LON)==0) {
      PrintOut(0, "Observer's coordinates must be specified.");
      return MPK_INVALID_PAR;
	} else {
	  cmpack_lontostr(ls_lon, buf1);
	  cmpack_lattostr(ls_lat, buf2);
  	  sprintf(msg,"Observer's coordinates are LON.=%s and LAT.=%s.", buf1, buf2);
  	  PrintOut(1, msg);
	}
  }
  
  switch (ls_format) 
  {
  case MPK_FORMAT_DIFFMAG:
    if (ls_nvar!=1) {
      PrintOut(0, "Exactly one variable star must be selected.");
	  return MPK_INVALID_PAR;
    }
    if (ls_ncmp<1) {
      PrintOut(0, "At least one comparison star must be selected.");
	  return MPK_INVALID_PAR;
    }
    ls_nstar = ls_nvar+ls_ncmp+ls_nchk;
    ls_mag   = (double *)malloc(ls_nstar*sizeof(double));
    ls_err   = (double *)malloc(ls_nstar*sizeof(double));
    n        = 2 + ls_nchk;
    ls_ncols = (n*(n-1))/2;
    ls_col   = (struct ColInfo *)malloc(ls_ncols*sizeof(struct ColInfo));
	for (i=0, k=0; i<n; i++) {
		for (j=i+1; j<n; j++, k++) {
			starid(buf1,i);
	  		starid(buf2,j);
	  		sprintf(ls_col[k].valname, "%s-%s", buf1, buf2);
	  		sprintf(ls_col[k].errname, "s%d", k+1);
       	}
	}
	ls_maxst = 0;
	for (st=ls_var; st!=NULL; st=st->next) {
    	if (st->index>ls_maxst) ls_maxst = st->index;
	}
	for (st=ls_cmp; st!=NULL; st=st->next) {
    	if (st->index>ls_maxst) ls_maxst = st->index;
	}
	for (st=ls_chk; st!=NULL; st=st->next) {
    	if (st->index>ls_maxst) ls_maxst = st->index;
	}
    break;

  case MPK_FORMAT_INSTMAG:
    ls_nstar = ls_nvar + ls_ncmp + ls_nchk;
    if (ls_nstar<1) {
      PrintOut(0, "At least one star must be selected.");
	  return MPK_INVALID_PAR;
    }
    ls_mag   = (double *)malloc(ls_nstar*sizeof(double));
    ls_err   = (double *)malloc(ls_nstar*sizeof(double));
    ls_ncols = ls_nstar;
    ls_col   = (struct ColInfo *)malloc(ls_ncols*sizeof(struct ColInfo));
	for (i=0; i<ls_nstar; i++) {
		sprintf(ls_col[i].valname, "MAG%d", i);
		sprintf(ls_col[i].errname, "ERR%d", i);
	}
	ls_maxst = 0;
	for (st=ls_var; st!=NULL; st=st->next) {
    	if (st->index>ls_maxst) ls_maxst = st->index;
	}
    break;

  case MPK_FORMAT_READALL:
  	break;

  case MPK_FORMAT_TRACKLST:
    ls_ncols = 2;
    ls_col   = (struct ColInfo *)malloc(ls_ncols*sizeof(struct ColInfo));
	sprintf(ls_col[0].valname, "OFFSETX");
	sprintf(ls_col[1].valname, "OFFSETY");
    break;
    
  case MPK_FORMAT_STDDEV:
    if (ls_ncmp<1) {
      PrintOut(0, "At least one comparison star must be selected.");
	  return MPK_INVALID_PAR;
    }
    if (ls_nchk<1) {
      PrintOut(0, "At least one check star must be selected.");
	  return MPK_INVALID_PAR;
    }
    ls_nstar = ls_ncmp+ls_nchk;
    ls_mag   = (double *)malloc(ls_nstar*sizeof(double));
    ls_err   = (double *)malloc(ls_nstar*sizeof(double));
    n        = ls_nchk + 1;
    ls_ncols = (n*(n-1))/2;
    ls_col   = (struct ColInfo *)malloc(ls_ncols*sizeof(struct ColInfo));
	for (i=0, k=0; i<n; i++) {
		for (j=i+1; j<n; j++, k++) {
			starid(buf1,i);
	  		starid(buf2,j);
	  		sprintf(ls_col[k].valname, "D(%s-%s)", buf1, buf2);
			sprintf(ls_col[k].errname, "E(%s-%s)", buf1, buf2);
       	}
	}
	ls_maxst = 0;
	for (st=ls_cmp; st!=NULL; st=st->next) {
    	if (st->index>ls_maxst) ls_maxst = st->index;
	}
	for (st=ls_chk; st!=NULL; st=st->next) {
    	if (st->index>ls_maxst) ls_maxst = st->index;
	}
	for (i=0; i<MAXAP; i++) {
		// Create new data row
		struct RowData *it = (struct RowData *)malloc(sizeof(struct RowData));
		memset(it, 0, sizeof(struct RowData));
		it->frame = i;
		it->ncols = ls_ncols;
		it->sum = (double *) malloc(ls_ncols*sizeof(double));
		memset(it->sum, 0, ls_ncols*sizeof(double));
		it->sum2 = (double *) malloc(ls_ncols*sizeof(double));
		memset(it->sum2, 0, ls_ncols*sizeof(double));
		it->count = (int *) malloc(ls_ncols*sizeof(int));
		memset(it->count, 0, ls_ncols*sizeof(int));
		if (ls_last==NULL) {
			ls_first = ls_last = it;
		} else {
			ls_last->next = it;
			ls_last = it;
		}
	}
    break;

  default:
	  sprintf(msg, "Invalid output format specification (%d)", ls_format);
	  PrintOut(0, msg);
	  return MPK_INVALID_PAR;
  }

  // Normal return
  if (alon) free(alon);
  if (alat) free(alat);
  return 0;
}

int mlist_read(const char *infile, int frame)
// Reads magnitudes and error of the selected stars and writes them to output file
{
	FILE *f;
	char buf[MAXLINE], msg[MAXLINE], *aux, *end;
	double jd,offsetx,offsety, ls, rs, vmag, verr, derr, cmag, dmag, cval, cerr;
	int j,k,n,delta,cok,vok;
	struct RowData *it;
	struct ColData *first_col, *next_col, *last_col, *col;
	struct StarData *st, *st2, *in_st;
	
	// Prints name of input file
	PrintOut(0, infile);
	
	// Read input file
	f = fopen(infile,"r");
	if (f==NULL)
		return MPK_CANNOT_OPEN_SRC;
	
	// Check file format
	fgets(buf,MAXLINE,f);
	if ((!strstr(buf," NL"))||(!strstr(buf,"JD")))
		return MPK_UNKNOWN_FORMAT;

	// Read header information
	fgets(buf,MAXLINE,f);
	jd = atof(buf+62);
	sprintf(msg, "JD = %.6f", jd);
	PrintOut(1, msg);
	if (strlen(ls_filter)==0)
		sscanf(buf+78,"%s",ls_filter);
  	  
	if (strlen(buf)>=132) {
		aux = buf+115;
		offsetx = strtod(aux,&end);
		if (end==aux) offsetx = 0.0;
		aux = buf+124;
		offsety = strtod(aux,&end);
		if (end==aux) offsety = 0.0;
	} else {
		offsetx = 0.0;
		offsety = 0.0;
	}

	fgets(buf,200,f);
	j = 3;

	switch (ls_format) 
	{
	case MPK_FORMAT_DIFFMAG:
		// Init magnitude and error arrays
		for (st=ls_var; st!=NULL; st=st->next) 
			memset(st->xvalid, 0, MAXAP*sizeof(int)); 
		for (st=ls_cmp; st!=NULL; st=st->next) 
			memset(st->xvalid, 0, MAXAP*sizeof(int)); 
		for (st=ls_chk; st!=NULL; st=st->next)
			memset(st->xvalid, 0, MAXAP*sizeof(int)); 
		// Read magnitudes and errors of stars
		while (fgets(buf,200,f) && (j<=ls_maxst*3)) {
			j++;
			// Convert index of line to index in magnitude and error arrays
			in_st = NULL; 
			for (st=ls_var; st!=NULL; st=st->next) {
				if (st->index*3+1==j) { in_st=st; break; }
			}
			if (!in_st) {
				for (st=ls_cmp; st!=NULL; st=st->next) {
					if (st->index*3+1==j) { in_st=st; break; }
				}
			}
			if (!in_st) {
				for (st=ls_chk; st!=NULL; st=st->next) {
					if (st->index*3+1==j) { in_st=st; break; }
				}
			}
			// Read magnitude
			if (in_st) {
				delta=25+9*(ls_aper-1);
				if ((int)strlen(buf)>delta) {
					aux = buf + delta;
					dmag = strtod(aux,&end);
					if (end!=aux && dmag<99 && dmag>-99) {
						in_st->xmag[0] = dmag;
						in_st->xvalid[0] |= VALID_MAG;
					}
				}
			}
			fgets(buf,200,f);
			j++;
			// Read error
			if (in_st) {
				delta=26+9*(ls_aper-1);
				if ((int)strlen(buf)>delta) {
					aux = buf + delta;
					derr = strtod(aux,&end);
					if (end!=aux && derr>0 && derr<9) {
						in_st->xerr[0] = derr;
						in_st->xvalid[0] |= VALID_ERR; 
					}
				}
			}
			// Skip empty line
			fgets(buf,200,f);
			j++;
			/*if (in_st) {
				sprintf(msg, "Star #%d at line #%d: %.3f %.3f %d", in_st->index, j-2, in_st->xmag[0], in_st->xerr[0], in_st->xvalid[0]);
				PrintOut(1, msg);
			}*/
		}

		// Create new data row
		it = (struct RowData *)malloc(sizeof(struct RowData));
		memset(it, 0, sizeof(struct RowData));
		it->frame = frame;
		it->ncols = ls_ncols;
		it->val = (double *) malloc(ls_ncols*sizeof(double));
		memset(it->val, 0, ls_ncols*sizeof(double));
		it->err = (double *) malloc(ls_ncols*sizeof(double));
		memset(it->err, 0, ls_ncols*sizeof(double));
		it->ok = (int *) malloc(ls_ncols*sizeof(int));
		memset(it->ok, 0, ls_ncols*sizeof(int));
		if (ls_last==NULL) {
			ls_first = ls_last = it;
		} else {
			ls_last->next = it;
			ls_last = it;
		}
		
		// Compute julian date of observation, make corrections
		it->jd = jd;
		if (ls_jdhel) {
			cmpack_sun(jd, &ls, &rs);
			it->jd += cmpack_hcor(ls_la, ls_be, ls, rs);
		} 
		if (ls_helcor) { 
			cmpack_sun(jd, &ls, &rs);
			it->hcor = cmpack_hcor(ls_la, ls_be, ls, rs);
		}
		if (ls_amass)  {
			it->amass = cmpack_airmass(jd, ls_ra, ls_dec, ls_lon, ls_lat);
			if (it->amass<=0) 
				PrintOut(0, "The object was under the horizon or its coordinates are bad.");
		}

		// Check data
		st = ls_var;
		if (!(st->xvalid[0] & VALID_MAG) || !(st->xvalid[0] & VALID_ERR)) {
			vmag = verr = 0.0;
			vok = 0;
			PrintOut(0, "The variable star has incorrect value of mag. or std. err.");
		} else {
			vmag = st->xmag[0];
			verr = st->xerr[0];
			vok  = VALID_MAG | VALID_ERR;
			//sprintf(msg, "Variable star: %.3f +- %.3f mag", vmag, verr);
			//PrintOut(1, msg);
		}
		
		// Compute artifitial comparison star magnitude and error
		cval = 0.0; cerr = 0.0; n = 0;
		for (st=ls_cmp; st!=NULL; st=st->next) {
			if ((st->xvalid[0] & VALID_MAG)!=0 && (st->xvalid[0] & VALID_ERR)!=0) {
				cval += pow(10.0, -0.4*st->xmag[0]);
				cerr += st->xerr[0];
				n++;
			}
		}
		if (n<1) {
			cmag = cerr = 0.0;
			cok = 0;
			PrintOut(0, "The comparison star(s) have incorrect value of mag. or std. err.");
		} else {
			cmag = -2.5*log10(cval/n);
			cerr = (cerr/n)/sqrt(n);
			cok = VALID_MAG | VALID_ERR;
			//sprintf(msg, "Comparison star: %.3f +- %.3f mag", cmag, cerr);
			//PrintOut(1, msg);
		}
		// First column is V-C
		if (vok && cok) {
			it->val[0] = vmag - cmag;
			it->err[0] = sqrt(verr*verr + cerr*cerr);
			it->ok[0] = VALID_MAG | VALID_ERR;
		} else {
			it->val[0] = it->err[0] = 0.0;
			it->ok[0] = 0;
		}
		// V-C1, V-C2, ...
		for (st=ls_chk, k=1; st!=NULL; st=st->next, k++) {
			if (vok && (st->xvalid[0] & VALID_MAG)!=0 && (st->xvalid[0] & VALID_ERR)!=0) {
				it->val[k] = vmag - st->xmag[0];
				it->err[k] = sqrt(verr*verr + st->xerr[0]*st->xerr[0]);
				it->ok[k]  = VALID_MAG | VALID_ERR;
			} else {
				it->val[k] = it->err[k] = 0.0;
				it->ok[k]  = 0;
			}
		}
		// C-C1, C-C2, ...
		for (st=ls_chk; st!=NULL; st=st->next, k++) {
			if (cok && (st->xvalid[0] & VALID_MAG)!=0 && (st->xvalid[0] & VALID_ERR)!=0) {
				it->val[k] = cmag - st->xmag[0];
				it->err[k] = sqrt(cerr*cerr + st->xerr[0]*st->xerr[0]);
				it->ok[k]  = VALID_MAG | VALID_ERR;
			} else {
				it->val[k] = it->err[k] = 0.0;
				it->ok[k]  = 0;
			}
		}
		// C1-C2, ...
		for (st=ls_chk; st!=NULL; st=st->next) {
			for (st2=st->next; st2!=NULL; st2=st2->next, k++) {
				if ((st->xvalid[0] & VALID_MAG)!=0 && (st->xvalid[0] & VALID_ERR)!=0 &&
				    (st2->xvalid[0] & VALID_MAG)!=0 && (st2->xvalid[0] & VALID_ERR)!=0) {
					it->val[k] = st->xmag[0] - st2->xmag[0];
					it->err[k] = sqrt(st->xerr[0]*st->xerr[0] + st2->xerr[0]*st2->xerr[0]);
					it->ok[k]  = VALID_MAG | VALID_ERR;
				} else {
					it->val[k] = it->err[k] = 0.0;
					it->ok[k]  = 0;
				}
			}
		}
		break;

	case MPK_FORMAT_INSTMAG:
	    // Init magnitude and error arrays
		for (st=ls_var; st!=NULL; st=st->next) 
			memset(st->xvalid, 0, MAXAP*sizeof(int)); 
		for (st=ls_cmp; st!=NULL; st=st->next) 
			memset(st->xvalid, 0, MAXAP*sizeof(int)); 
		for (st=ls_chk; st!=NULL; st=st->next)
			memset(st->xvalid, 0, MAXAP*sizeof(int)); 
        // Read magnitudes and errors of stars
        while (fgets(buf,200,f) && (j<=ls_maxst*3)) {
			j++;
			// Convert index of line to index in magnitude and error arrays
			in_st = NULL; 
			for (st=ls_var; st!=NULL; st=st->next) {
				if (st->index*3+1==j) { in_st=st; break; }
			}
			if (!in_st) {
				for (st=ls_cmp; st!=NULL; st=st->next) {
					if (st->index*3+1==j) { in_st=st; break; }
				}
			}
			if (!in_st) {
				for (st=ls_chk; st!=NULL; st=st->next) {
					if (st->index*3+1==j) { in_st=st; break; }
				}
			}
			// Read magnitude
			if (in_st) {
				delta=25+9*(ls_aper-1);
				if ((int)strlen(buf)>delta) {
					aux = buf + delta;
					dmag = strtod(aux,&end);
					if (end!=aux && dmag<99 && dmag>-99) {
						in_st->xmag[0] = dmag;
						in_st->xvalid[0] |= VALID_MAG;
					}
				}
			}
			fgets(buf,200,f);
			j++;
			// Read error
			if (in_st) {
				delta=26+9*(ls_aper-1);
				if ((int)strlen(buf)>delta) {
					aux = buf + delta;
					derr = strtod(aux,&end);
					if (end!=aux && derr>0 && derr<9) {
						in_st->xerr[0] = derr;
						in_st->xvalid[0] |= VALID_ERR; 
					}
				}
			}
			// Skip empty line
			fgets(buf,200,f);
			j++;
			/*if (in_st) {
				sprintf(msg, "Star #%d at line #%d: %.3f %.3f", in_st->index, j-2, in_st->xmag[0], in_st->xerr[0]);
				PrintOut(1, msg);
			}*/
		}
		// Create new data row
		it = (struct RowData *)malloc(sizeof(struct RowData));
		memset(it, 0, sizeof(struct RowData));
		it->frame = frame;
		it->ncols = ls_ncols;
		it->val = (double *) malloc(ls_ncols*sizeof(double));
		memset(it->val, 0, ls_ncols*sizeof(double));
		it->err = (double *) malloc(ls_ncols*sizeof(double));
		memset(it->err, 0, ls_ncols*sizeof(double));
		it->ok = (int *) malloc(ls_ncols*sizeof(int));
		memset(it->ok, 0, ls_ncols*sizeof(int));
		if (ls_last==NULL) {
			ls_first = ls_last = it;
		} else {
			ls_last->next = it;
			ls_last = it;
		}
		// Compute julian date of observation, make corrections
		it->jd = jd;
		if (ls_jdhel) {
			cmpack_sun(jd, &ls, &rs);
			it->jd += cmpack_hcor(ls_la, ls_be, ls, rs);
		} 
		if (ls_helcor) { 
			cmpack_sun(jd, &ls, &rs);
			it->hcor = cmpack_hcor(ls_la, ls_be, ls, rs);
		}
		if (ls_amass)  {
			it->amass = cmpack_airmass(jd, ls_ra, ls_dec, ls_lon, ls_lat);
			if (it->amass<=0) 
				PrintOut(0, "The object was under the horizon or its coordinates are bad.");
		}
		// Copy instrumental magnitudes and errors
		k = 0;
		for (st=ls_var; st!=NULL; st=st->next) {
			if ((st->xvalid[0] & VALID_MAG)!=0 && (st->xvalid[0] & VALID_ERR)!=0) {
				it->val[k] = st->xmag[0];
				it->err[k] = st->xerr[0];
				it->ok[k]  = VALID_MAG | VALID_ERR;
			} else {
				it->ok[k]  = 0;
			}
			k++;
		}
		for (st=ls_cmp; st!=NULL; st=st->next) {
			if ((st->xvalid[0] & VALID_MAG)!=0 && (st->xvalid[0] & VALID_ERR)!=0) {
				it->val[k] = st->xmag[0];
				it->err[k] = st->xerr[0];
				it->ok[k]  = VALID_MAG | VALID_ERR;
			} else {
				it->ok[k]  = 0;
			}
			k++;
		}
		for (st=ls_chk; st!=NULL; st=st->next) {
			if ((st->xvalid[0] & VALID_MAG)!=0 && (st->xvalid[0] & VALID_ERR)!=0) {
				it->val[k] = st->xmag[0];
				it->err[k] = st->xerr[0];
				it->ok[k]  = VALID_MAG | VALID_ERR;
			} else {
				it->ok[k]  = 0;
			}
			k++;
		}
      	break;

    case MPK_FORMAT_READALL:
      	// Read magnitudes and errors of stars
      	first_col = last_col = NULL; 
      	j = 0;
      	while (fgets(buf,MAXLINE-1,f)) {
          	j++;
	      	col = (struct ColData*) malloc(sizeof(struct ColData));
	      	memset(col, 0, sizeof(struct ColData));
          	// Read magnitude from the first line
          	delta=25+9*(ls_aper-1);
          	if ((int)strlen(buf)>delta) {
              	aux = buf + delta;
      	      	col->val = strtod(aux,&end);
              	if (end!=aux && col->val<99 && col->val>-99)
              		col->ok |= VALID_MAG;
          	}
          	fgets(buf,MAXLINE-1,f);
          	j++;
          	// Read error from the second line
          	delta=26+9*(ls_aper-1);
          	if ((int)strlen(buf)>delta) {
	           	aux = buf + delta;
	   			col->err = strtod(aux,&end);
       			if (end!=aux && col->err<9 && col->err>-9)
       				col->ok |= VALID_ERR;
          	}
          	if (last_col) {
	          	col->next = NULL;
	          	last_col->next = col;
	          	last_col = col;
          	} else {
	          	first_col = last_col = col;
          	}
          	// Skip next (empty) line
          	fgets(buf,MAXLINE-1,f);
          	j++;
      	}
      	j = j/3;
      	if (j>0) {
	      	// Create a new row record
  	  		it = (struct RowData *)malloc(sizeof(struct RowData));
  	  		memset(it, 0, sizeof(struct RowData));
  	  		it->frame = frame;
	  		it->val = (double *)malloc(j*sizeof(double));
	  		memset(it->val, 0, j*sizeof(double));
	  		it->err = (double *)malloc(j*sizeof(double));
	  		memset(it->err, 0, j*sizeof(double));
	  		it->ok  = (int *)malloc(j*sizeof(int));
	  		memset(it->ok, 0, j*sizeof(int));
      		if (ls_first==NULL) {
		  		ls_first = ls_last = it;
	  		} else {
		  		ls_last->next = it;
		  		ls_last = it;
	  		}
	  		// Compute julian date, make corrections
      		it->jd = jd;
  	  		if (ls_jdhel) {
 				cmpack_sun(jd, &ls, &rs);
 				it->jd += cmpack_hcor(ls_la, ls_be, ls, rs);
	  		}
	      	// Transform column data into row data
	  		it->ncols = 0;
	      	for (j=0, col=first_col; col!=NULL; col=col->next, j++) {
		  		it->val[j] = col->val;
		  		it->err[j] = col->err;
		  		it->ok[j]  = col->ok;
      			if (col->ok)
      				it->ncols = j+1;  // Number of valid columns
	  		}
	  		// Free column data
			col = first_col; 
	  		while (col) {
      			next_col = col->next;
      			free(col);
      			col = next_col;
	  		}
  	  	}
      	break;

    case MPK_FORMAT_TRACKLST:
      	// Create new data row
  	  	it = (struct RowData *)malloc(sizeof(struct RowData));
  	  	memset(it, 0, sizeof(struct RowData));
  	  	it->frame = frame;
  	  	it->ncols = ls_ncols;
	  	it->val = (double *)malloc(it->ncols*sizeof(double));
	  	memset(it->val, 0, it->ncols*sizeof(double));
	  	it->err = (double *)malloc(it->ncols*sizeof(double));
	  	memset(it->err, 0, it->ncols*sizeof(double));
	  	it->ok  = (int *)malloc(it->ncols*sizeof(int));
	  	memset(it->ok, 0, it->ncols*sizeof(int));
	  	if (ls_first==NULL) {
		  	ls_first = ls_last = it;
	  	} else {
		  	ls_last->next = it;
		  	ls_last = it;
	  	}
		// Compute julian date of observation, make corrections
		it->jd = jd;
		if (ls_jdhel) {
			cmpack_sun(jd, &ls, &rs);
			it->jd += cmpack_hcor(ls_la, ls_be, ls, rs);
		} 
      	// Copy instrumental magnitudes and errors
      	it->val[0] = offsetx;
      	it->ok[0] = VALID_MAG;
      	it->val[1] = offsety;
      	it->ok[1] = VALID_MAG;
      	break;
      	
	case MPK_FORMAT_STDDEV:
		// Init magnitude and error arrays
		for (st=ls_cmp; st!=NULL; st=st->next) {
			memset(st->xvalid, 0, MAXAP*sizeof(int));
		}
		for (st=ls_chk; st!=NULL; st=st->next) {
			memset(st->xvalid, 0, MAXAP*sizeof(int));
		}
		// Read magnitudes and errors of stars
		while (fgets(buf,200,f) && (j<=ls_maxst*3)) {
			j++;
			// Convert index of line to index in magnitude and error arrays
			in_st = NULL; 
			for (st=ls_cmp; st!=NULL && in_st==NULL; st=st->next) {
				if (st->index*3+1==j) in_st=st; 
			}
			for (st=ls_chk; st!=NULL && in_st==NULL; st=st->next) {
				if (st->index*3+1==j) in_st=st;
			}
			// Read magnitudes
			if (in_st) {
				for (k=0; k<MAXAP; k++) {
					delta=25+9*k;
					if ((int)strlen(buf)>delta) {
						aux = buf + delta;
						dmag = strtod(aux,&end);
						if (end!=aux && dmag<99 && dmag>-99) {
							in_st->xmag[k] = dmag;
							in_st->xvalid[k] |= VALID_MAG;
						}
					}
				}
			}
			// Skip errors
			fgets(buf,200,f);
			j++;
			// Read errors
			if (in_st) {
				for (k=0; k<MAXAP; k++) {
					delta=26+9*k;
					if ((int)strlen(buf)>delta) {
						aux = buf + delta;
						derr = strtod(aux,&end);
						if (end!=aux && derr>0 && derr<9) {
							in_st->xerr[k] = derr;
							in_st->xvalid[k] |= VALID_ERR; 
						}
					}
				}
			}
			// Skip empty line
			fgets(buf,200,f);
			j++;
			/*if (in_st) {
				sprintf(msg, "Star #%d at line #%d: %.3f +- %.3f", in_st->index, j-2, in_st->xmag[0], in_st->xerr[0]);
				PrintOut(1, msg);
			}*/
		}

		for (k=0, it=ls_first; k<MAXAP && it!=NULL; k++, it=it->next) {
			// Compute artifitial comparison star magnitude and error
			cval = 0.0; n = 0;
			for (st=ls_cmp; st!=NULL; st=st->next) {
				if ((st->xvalid[k] & VALID_MAG)!=0 && (st->xvalid[k] & VALID_ERR)!=0) {
					cval += pow(10.0, -0.4*st->xmag[k]);
					n++;
				}
			}
			if (n<1) {
				PrintOut(0, "The comparison star(s) have incorrect value of mag. or std. err.");
				cmag = 0.0;
				cok = 0;
			} else {
				cmag = -2.5*log10(cval/n);
				cok = VALID_MAG;
				//sprintf(msg, "Comparison star: %.3f", cmag);
				//PrintOut(1, msg);
			}
			j = 0;
			// C-C1, C-C2, ...
			for (st=ls_chk; st!=NULL; st=st->next) {
				if (cok && (st->xvalid[k] & VALID_MAG)!=0 && (st->xvalid[k] & VALID_ERR)!=0) {
					dmag = cmag - st->xmag[k];
					it->sum[j] += dmag;
					it->sum2[j] += dmag*dmag;
					it->count[j]++;
				}
				j++;
			}
			// C1-C2, ...
			for (st=ls_chk; st!=NULL; st=st->next) {
				for (st2=st->next; st2!=NULL; st2=st2->next) {
					if ((st->xvalid[k] & VALID_MAG)!=0 && (st->xvalid[k] & VALID_ERR)!=0 && 
						(st2->xvalid[k] & VALID_MAG)!=0 && (st2->xvalid[k] & VALID_ERR)!=0) {
						dmag = st->xmag[k] - st2->xmag[k];
						it->sum[j] += dmag;
						it->sum2[j] += dmag*dmag;
						it->count[j]++;
					}
					j++;
				}
			}
		}
		break;
  }

  // Normal return
  fclose(f);
  return 0;
}

int mlist_delete(int frame)
// Deletes specified frame from the list
{
    struct RowData *it = ls_first, *prev = NULL;
    while (it) {
	    if (it->frame == frame) {
		    if (prev==NULL) {
			    ls_first = it->next;
  			} else {
	  			prev->next = it->next;
  			}
  			if (it->val) 
  				free(it->val);
  			if (it->err)
  				free(it->err);
  			if (it->ok)
  				free(it->ok);
  			if (it->sum)
  				free(it->sum);
  			if (it->sum2)
  				free(it->sum2);
  			if (it->count)
  				free(it->count);
  			free(it);
  			return 0;
	    }
	    prev = it;
	    it = it->next;
    }
    return MPK_FRAME_NOT_FOUND;
}

int mlist_getnrows(void)
// Returns number of rows
{
    int nrows = 0;
    struct RowData *it;
    for (it=ls_first; it!=NULL; it = it->next)
	    nrows++;
    return nrows;
}

int mlist_getncols(void)
// Returns number of columns
{
	return ls_ncols;
}

int mlist_getcoli(int col, char *buf)
// Get name of the specified column
{
	if (col>=0 && col<ls_ncols) {
		if (buf) strcpy(buf, ls_col[col].valname);
		return 0;
	} else {
		if (buf) buf[0] = '\0';
		return MPK_INVALID_PAR;
	}
}

int mlist_getrowi(int row, int *frm, double *jd, double *hcor, double *amass)
// Returns light curve of a selected star
{
    struct RowData *it = ls_first;
    int i = 0;
    while (i<row && it) {
	    it = it->next;
	    i++;
    }
    switch (ls_format) 
    {
	case MPK_FORMAT_DIFFMAG:
	case MPK_FORMAT_INSTMAG:
	case MPK_FORMAT_TRACKLST:
	    if (it) {
		  	if (frm)   *frm   = it->frame;
		  	if (jd)    *jd    = it->jd;
		  	if (hcor)  *hcor  = it->hcor;
		  	if (amass) *amass = it->amass;
		  	return 0;
		} else {
		    if (frm)   *frm   = -1;
		    if (jd)    *jd    = 0.0; 
		    if (hcor)  *hcor  = 0.0;
		    if (amass) *amass = 0.0;
	    	return MPK_INVALID_PAR;
		}
		break;
	  
	case MPK_FORMAT_STDDEV:
	    if (it) {
		  	if (frm)   *frm   = it->frame;
		  	if (jd)    *jd    = it->frame+1;
		  	if (hcor)  *hcor  = 0.0;
		  	if (amass) *amass = 0.0;
		  	return 0;
		} else {
		    if (frm)   *frm   = -1;
		    if (jd)    *jd    = 0; 
		    if (hcor)  *hcor  = 0.0;
		    if (amass) *amass = 0.0;
	    	return MPK_INVALID_PAR;
		}
		break;
	}
	return MPK_INVALID_PAR;
}

int mlist_getval(int row, int col, double *val, double *err, int *valid)
// Returns light curve of a selected star
{
	int i;
    struct RowData *it;
    
    if (valid) 	*valid = 0;
	if (val) 	*val = 99.999;
	if (err) 	*err = 9.9999;
    
	if (col>=0 && col<ls_ncols) {
	    for (it=ls_first, i=0; it!=NULL; it=it->next, i++) {
		    if (i==row) {
			    switch (ls_format) 
			    {
				case MPK_FORMAT_DIFFMAG:
				case MPK_FORMAT_INSTMAG:
  				case MPK_FORMAT_TRACKLST:
					if ((it->ok[col] & VALID_MAG)!=0) {
						if (val)   *val = it->val[col];
						if (valid) *valid |= VALID_MAG;
					}
					if ((it->ok[col] & VALID_ERR)!=0) {
						if (err)   *err = it->err[col];
						if (valid) *valid |= VALID_ERR;
					}
					break;
					
				case MPK_FORMAT_STDDEV:
		    		if (it->count[col]>0) {
			    		if (it->count[col]>1) {
				    		double diff = it->count[col]*it->sum2[col] - it->sum[col]*it->sum[col];
				    		if (diff>0.0000001) {
				    			if (val) *val = (1.0/it->count[col])*sqrt(diff);
				    			if (valid) *valid |= VALID_MAG;
			    			} else {
				    			if (val) *val = 0.0;
				    			if (valid) *valid |= VALID_MAG;
			    			}
			    		} else {
			    			if (val) *val = 0.0;
			    			if (valid) *valid |= VALID_MAG;
		    			}
	    			}
					break;
				}
				return 0;
			}
		}
	}
	return MPK_INVALID_PAR;
}

int mlist_write(const char *outfile)
// Writes data to file
{
  FILE *outf;
  int i;
  double val, err;
  struct RowData *it;

  // Open output file
  outf = fopen(outfile,"w+");
  if (outf==NULL) return MPK_CANNOT_OPEN_OUT;

  switch (ls_format) {
	  case MPK_FORMAT_DIFFMAG:
	  case MPK_FORMAT_INSTMAG:
	    // First row: list of columns
	    fprintf(outf, "%s ", (ls_jdhel ? "JDHEL" : "JD"));
	    for (i=0; i<ls_ncols; i++)
		    fprintf(outf, "%s %s ", ls_col[i].valname, ls_col[i].errname);
	    if (ls_helcor) fprintf(outf,"HELCOR ");
	    if (ls_amass) fprintf(outf,"AIRMASS ");
        fprintf(outf,"\n");
        // Second row: parameters
		fprintf(outf,"Aperture: %d", ls_aper);
		if (strlen(ls_filter)>0) fprintf(outf,", Filter: %s",ls_filter);
		if (ls_jdhel) fprintf(outf,", JD: heliocentric");
		         else fprintf(outf,", JD: geocentric");
        fprintf(outf,"\n");
        // Print data
        for (it=ls_first; it!=NULL; it = it->next) {
			fprintf(outf,"%.5f ", it->jd);
    		for (i=0; i<it->ncols; i++) {
	    		val = ((it->ok[i] & VALID_MAG)!=0 ? it->val[i] : 99.999);
	    		err = ((it->ok[i] & VALID_ERR)!=0 ? it->err[i] : 9.999);
	    		fprintf(outf, "%.4f %.4f ", val, err);
    		}
    		if (ls_helcor)
    			fprintf(outf,"%.5f ", it->hcor);
    		if (ls_amass)
    			fprintf(outf,"%.3f ", it->amass);
        	fprintf(outf, "\n");
	    }
    	break;

      case MPK_FORMAT_TRACKLST:
        // Print header
        fprintf(outf,"# %s OFFSETX OFFSETY\n", (ls_jdhel ? "JDHEL" : "JD"));
        fprintf(outf,"\n");
        // Print data
        for (it=ls_first; it!=NULL; it = it->next) {
			fprintf(outf,"%10.5f %8.3f %8.3f", it->jd, it->val[0], it->val[1]);
    		fprintf(outf, "\n");
	    }
        break;

      case MPK_FORMAT_READALL: 
        // Print header
        fprintf(outf,"# %s, instrumental mags and standard deviations of all detected stars\n", (ls_jdhel ? "JDHEL" : "JD"));
	    fprintf(outf,"# Aperture: %d",ls_aper);
	    if (strlen(ls_filter)>0) fprintf(outf,", Filter: %s",ls_filter);
		if (ls_jdhel) fprintf(outf,", JD: heliocentric");
		         else fprintf(outf,", JD: geocentric");
        fprintf(outf,"\n");
      	// Allstars output format
        for (it=ls_first; it!=NULL; it = it->next) {
			fprintf(outf,"%.5f ", it->jd);
    		for (i=0; i<it->ncols; i++) {
	    		val = ((it->ok[i] & VALID_MAG)!=0 ? it->val[i] : 99.999);
	    		err = ((it->ok[i] & VALID_ERR)!=0 ? it->err[i] : 9.999);
	    		fprintf(outf, "%.4f %.4f ", val, err); 
	    	}
        	fprintf(outf, "\n");
    	}
        break;
        
	  case MPK_FORMAT_STDDEV:
	    // First row: list of columns
	    fprintf(outf, "AP# ");
	    for (i=0; i<ls_ncols; i++)
		    fprintf(outf, "%s ", ls_col[i].valname);
        fprintf(outf,"\n");
        // Second row: parameters
		if (strlen(ls_filter)>0) 
			fprintf(outf,"Filter: %s", ls_filter);
        fprintf(outf,"\n");
        // Print data
        for (it=ls_first; it!=NULL; it = it->next) {
			fprintf(outf,"%d ", it->frame+1);
    		for (i=0; i<it->ncols; i++) {
	    		if (it->count[i]>0) {
		    		if (it->count[i]>1) {
			    		double diff = it->count[i]*it->sum2[i] - it->sum[i]*it->sum[i];
			    		if (diff>0.0000001)
			    			val = (1.0/it->count[i])*sqrt(diff);
			    		else
			    			val = 0.0;
		    		} else {
			    		val = 0.0;
		    		}
	    		} else {
		    		val = 99.999;
	    		}
	    		fprintf(outf, "%.4f ", val);
    		}
        	fprintf(outf, "\n");
	    }
    	break;
  }

  fclose(outf);
  // Normal return
  return 0;
}

int mlist_clean()
// Closes output file and frees allocated memory
{
	struct StarData *st, *next;  
	struct RowData *ptr;

	free(ls_mag);
	free(ls_err);
	free(ls_col);
	ls_mag = ls_err = NULL;
	ls_col = NULL;
	ls_nstar = ls_ncols = 0;

	st = ls_var;
	while (st) { 
		next = st->next; 
		free(st); 
		st = next; 
	}
	st = ls_cmp;
	while (st) { 
		next = st->next; 
		free(st); 
		st = next; 
	}
	st = ls_chk;
	while (st) { 
		next = st->next; 
		free(st); 
		st = next; 
	}
	ls_var = ls_cmp = ls_chk = NULL;
	ls_var_last = ls_cmp_last = ls_chk_last = NULL;
	ls_nvar = ls_ncmp = ls_nchk = 0;
	
	ptr = ls_first;
	while (ptr) {
		struct RowData *next = ptr->next;
		if (ptr->val)
			free(ptr->val);
		if (ptr->err)
			free(ptr->err);
		if (ptr->ok)
			free(ptr->ok);
		free(ptr);
		ptr = next;
	}
	ls_first = ls_last = NULL;
	return 0;
}
