/**************************************************************

munilist.c (C-Munipack project)
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <math.h>

#ifdef _WIN32
#include <io.h>
#endif

#include "comfun.h"
#include "trajd.h"
#include "mlist.h"

void PrintHelp()
{
 printf("MUNILIST version %s - listing magnitude and time from photometry files.\n",VERSION);
 printf(PROJECTLINE);
 printf("\nUsage: munilist [options] [@dirfile] [var=n] [comp=n,...] [chk=n,...] [image1 ...]\n\n");
 printf("parameters:\n");
 printf("     @dirfile    file with input image names\n");
 printf("                 (for only '@', read filenames from standart input)\n");
 printf("     var=n       index of line with variable star\n");
 printf("     comp=n,...  indices of lines with comparison stars\n");
 printf("     chk=n,...   indices of lines with check stars\n");
 printf("     par=file    parameter file (default is %s)\n",OPTFILE);
 printf("     col=n       column for listing\n");
 printf("     ap=n        aperture number\n");
 printf("     out=file    output file (default = %s)\n",OUTFILE);
 printf("     ra=         right ascension of the object \n");
 printf("     dec=        declination of the object \n");
 printf("                 (the object's coordinates are required for JD(hel) or HelCor computation)\n");
 printf("     lon=        geographical longitude of the observer\n");
 printf("     lat=        geographical latitude of the observer\n");
 printf("                 (the observer's coordinates are required for airmass computation)\n");
 printf("     image1...   names of simple scientific exposures\n");
 printf("options:\n");
 printf("     -p<file>    parameter file (equivalent with 'par=')\n");
 printf("     -q          quiet mode\n");
 printf("     -h          give this help\n");
 printf("     -v          prints software version\n");
 printf("     -l          display software license\n");
 printf("     -g          creates the example of the parameter file\n");
 printf("output format selection (only one of the following):\n");
 printf("     -os         differential magnitudes\n");
 printf("     -of         instrumental magnitudes\n");
 printf("     -oa         all-stars (readall format)\n");
 printf("     -ot         track list\n");
 printf("     -od         standard deviations\n");
 printf("output format modifiers (combination of the following):\n");
 printf("     -jdhel      print heliocentric JD\n");
 printf("     -helcor     include heliocentric correction\n");
 printf("     -airmass    include airmass coefficient\n");
 printf(WARRANTYLINE);
 printf(HOMEPAGELINE);
 printf(BUGREPORTS);
}

void report(int status, const char *file)
{
 if (cmpack_debug) 
 	fprintf(stderr, "The following error(s) has been occurred: %d\n",status);
 PrintError(PROGNAME, file, 0, status);
}

void error(int status, const char *file)
{
 report(status, file);
 exit(status);
}

void CreatePar()
/* Creates a sample parameter file */
{
 FILE *fp;

 if ((fp=fopen(SMPFILE,"w"))!=NULL) {
   printf("Writing '%s' ...\n\n",SMPFILE);
   fprintf(fp,"APERTURE = %d     # Aperture number\n",DEFAPERTURE);
   fprintf(fp,"OUTFORM  = %d     # Output format\n",DEFOUTFORM);
   fprintf(fp,"JDFORM   = 0      # Julian date format\n");
   fprintf(fp,"HELCOR   = 0      # Include heliocentric correction\n");
   fprintf(fp,"AIRMASS  = 0      # Include airmass coefficient\n");
   fprintf(fp,"LON      = 0      # Observer's longitude\n");
   fprintf(fp,"LAT      = 0      # Observer's latitude\n");
   fclose(fp);
   printf("'%s' is an example of the parameter file. Rename it to '%s' and edit the parameters.\n\n",SMPFILE,OPTFILE);
 }
}

static char *AddPar(char *list, const char *par)
/* Add parameter to the parameter list */
{
	if (list) {
		list = realloc(list, strlen(list)+strlen(par)+2);
		strcat(list, " ");
		strcat(list, par);
	} else {
		list = malloc(strlen(par)+1);
		strcpy(list, par);
	}
	return list; 
}

int main(int argc, char **argv)
{
    int i, n, res, stdinput = 0, format = -1, ap = -1, jdhel=0, helcor=0, airmass=0, level = 0;
    char line[MAXLINE], buf1[MAXLINE], *par, *parlist = NULL;
	char *namsci = "", *namopt = NULL, *namout = OUTFILE;
	FILE *f;

	/* Zpracovani parametru prikazove radky */
	for (i=1;i<argc;i++) {
	   par = argv[i];
	   /* Prints help */
	   if ((strcmp(par,"-h")==0)||(strcmp(par,"--help")==0)) {
		  PrintHelp();
          return 0;
       }
	   /* Prints version */
	   if ((strcmp(par,"-v")==0)||(strcmp(par,"--version")==0)) {
		  PrintVersion(PROGNAME);
          return 0;
       }
       /* Prints license */
       if ((strcmp(par,"-l")==0)||(strcmp(par,"--license")==0)) {
          PrintLicense();
          return 0;
       }
       /* Generate sample parameter file */
       if (strcmp(par,"-g")==0) {
           CreatePar();
           return 0;
       }
       /* Flag for input from stdin or dirfile */
       if (strchr(par,'@')==par) {
	       namsci = malloc(strlen(par));
	       strcpy(namsci,&par[1]);
           stdinput = strlen(namsci)==0;
       }
       /* Aperture selection */
       if (strstr(par,"col=")==par) {
	       ap = (atol(par+4)-2)/2;
	       if (ap<=0) ap = 1;
       }
       if (strstr(par,"ap=")==par) {
	       ap = atol(par+3);
	       if (ap<=0) ap = 1;
       }
       /* Output files */
       if (strstr(par,"out=")==par) {
	       namout = malloc(strlen(par)-3);
	       strcpy(namout,&par[4]);
       }
       /* Parameter file */
       if (strstr(par,"par=")==par) {
	       namopt = malloc(strlen(par)-3);
	       strcpy(namopt,&par[4]);
       }
       if (strstr(par,"-p")==par) {
	       namopt = malloc(strlen(par)-1);
	       strcpy(namopt,&par[2]);
       }
       /* Output format selection */
       if (strcmp(par,"-os")==0)
	       format = MPK_FORMAT_DIFFMAG;
       if (strcmp(par,"-of")==0)
	       format = MPK_FORMAT_INSTMAG;
       if (strcmp(par,"-oa")==0)
           format = MPK_FORMAT_READALL;
       if (strcmp(par,"-ot")==0)
	       format = MPK_FORMAT_TRACKLST;
       if (strcmp(par,"-od")==0)
	       format = MPK_FORMAT_STDDEV;
       /* Output format modifiers */
       if (strcmp(par,"-jdhel")==0)
	       jdhel = 1;
       if (strcmp(par,"-helcor")==0)
	       helcor = 1;
       if (strcmp(par,"-airmass")==0)
	       airmass = 1;
       /* Object's coordinates */
       if (strstr(par,"ra=")==par || strstr(par,"dec=")==par) {
	       parlist = AddPar(parlist, par);
       }
       /* Observer's coordinates */
       if (strstr(par,"lon=")==par || strstr(par,"lat=")==par) {
	       parlist = AddPar(parlist, par);
       }
       /* Selected stars */
       if (strstr(par,"var=")==par || strstr(par,"comp=")==par || strstr(par,"chk=")==par) {
	       parlist = AddPar(parlist, par);
       }
       /* Quiet/verbose flags */
       if (strcmp(par,"-q")==0 || strcmp(par,"--quiet")==0 || strcmp(par,"--silent")==0) {
           level = MPK_LEVEL_QUIET;
       }
       if (strcmp(par,"--verbose")==0 || strcmp(par,"--debug")==0) {
           level = MPK_LEVEL_DEBUG;
       }
  }
  
  cmpack_setoutput(PrintMessage, level);
  PrintPrologue(PROGNAME);
  format = format | (jdhel?MPK_FORMAT_JDHEL:0) | (helcor?MPK_FORMAT_HELCOR:0) | (airmass?MPK_FORMAT_AIRMASS:0);
  res = mlist_init(namopt, ap, format, parlist);
  if (res!=0) error(res, NULL);
  n = 0;

  /* image names reads from command line */
  for (i=1;i<argc;i++) {
	  par = argv[i];
      if ((strchr(par,'=')==0)&&(strchr(par,'@')!=par)&&(strchr(par,'-')!=par)&&(atol(par)==0)) {
#ifdef _WIN32
		  intptr_t fh;
		struct _finddata_t fd;
		 if ((fh=_findfirst(par, &fd))!=-1) {
			do {
				if ((fd.attrib & _A_SUBDIR)==0)  {
          			res=mlist_read(fd.name, n+1);
          			if (res!=0) report(res, fd.name);
          			n++;
          		}
          	} while (_findnext(fh, &fd)==0);
			_findclose(fh);
      	 } else {
	      	 report(MPK_CANNOT_OPEN_SRC, par);
      	 }
#else
		res=mlist_read(par, n+1);
		if (res!=0) report(res, par);
		n++;
#endif
      }
  }

  /* image names reads from dirfile */
  if (strlen(namsci)>0) {
	  f = fopen(namsci,"r");
	  if (f==NULL) error(MPK_NO_DIRFILE, namsci);
      while(fgets(buf1,MAXLINE,f)!=NULL) {
          sscanf(buf1,"%s",line);
          if (strcmp(line,"")) {
	          res=mlist_read(line,n+1);
    	      if (res) report(res, line);
	          n++;
          }
      }
  }

  /* Image names reads from stdin */
  if (stdinput) {
      while(fgets(buf1,MAXLINE,stdin)!=NULL) {
          sscanf(buf1,"%s",line);
          if (strcmp(line,"")) {
	          res=mlist_read(line,n+1);
    	      if (res) report(res, line);
	          n++;
          }
      }
  }

  if (n<=0) 
      PrintOut(0, "No input files.");

  /* Create output file */
  if (res==0) {
      res = mlist_write(namout);
      if (res!=0) report(res, namout);
  }

  /* Delete parameter list */  
  if (parlist)
	  free(parlist);

  /* Normal return */
  res = mlist_clean();
  if (res!=0) error(res, NULL);

  return 0;
}
