/**************************************************************

match.c (C-Munipack project)
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#define _NO_OLDNAMES
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <math.h>

#include "comfun.h"
#include "inifile.h"
#include "match.h"
#include "matfun.h"
#include "matread.h"
#include "matwrite.h"
#include "matsolve.h"
#include "matsimple.h"

int match_init(const char *reffile, const char *optfile)
{
  inifile *f;
  char msg[MAXLINE];
  int res, max2;

  /* Read parameter file */
  f = iniopen(optfile, OPTFILE);
  if (f==NULL) {
		if (optfile) sprintf(msg,"Cannot read parameter file '%s'",optfile);
		        else sprintf(msg,"Cannot read parameter file '%s'",OPTFILE);
		PrintOut(0, msg);
		return MPK_CANNOT_OPEN_PAR;
	}
	mmatch_maxstar   = inigkyl(f, "RSTARS", DEFRSTARS);
	mmatch_nstar     = inigkyl(f, "ISTARS", DEFISTARS);
	mmatch_clip      = inigkyd(f, "CLIP", DEFCLIP);
	mmatch_spfields  = inigkyl(f, "SP_FIELDS", 0);
	mmatch_maxoffset = inigkyd(f, "MAXOFFSET", 5.0);
	iniclos(f);

  /* Check parameters */
  if (mmatch_spfields<0 || mmatch_spfields>2) 
      mmatch_spfields = 0;
      
  if (mmatch_nstar<=2) {
	  PrintOut(0, "Number of identification stars muse be greater than 2");
	  return MPK_INVALID_PAR;
  }
  if (mmatch_nstar>=20) {
	  PrintOut(0, "Number of identification stars muse be less than 20");
	  return MPK_INVALID_PAR;
  }	  
  if (mmatch_maxstar<mmatch_nstar) {
	  PrintOut(0, "Number of stars used muse be greater or equal to number of identification stars");
	  return MPK_INVALID_PAR;
  }
  if (mmatch_maxstar>=1000) {
	  PrintOut(0, "Number of stars used for matching muse be less than 1000");
	  return MPK_INVALID_PAR;
  }
  if (mmatch_clip<=0) {
	  PrintOut(0, "Clipping factor must be greater than zero");
	  return MPK_INVALID_PAR;
  }
  if (mmatch_maxoffset<=0) {
	  PrintOut(0, "Max. position offset muse be greater than zero");
	  return MPK_INVALID_PAR;
  }

  /* Allocate memory buffers */
  mmatch_idr1 = (int*) malloc(mmatch_maxstar*sizeof(int));
  if (!mmatch_idr1) 
      return MPK_MEMORY;
  mmatch_idr2 = (int*) malloc(mmatch_maxstar*sizeof(int));
  if (!mmatch_idr2)
      return MPK_MEMORY;
  mmatch_id1 = (int*) malloc(mmatch_maxstar*sizeof(int));
  if (!mmatch_id1)
      return MPK_MEMORY;
  mmatch_id2 = (int*) malloc(mmatch_maxstar*sizeof(int));
  if (!mmatch_id2)
      return MPK_MEMORY;
  mmatch_r1 = (double*) malloc(mmatch_maxstar*sizeof(double));
  if (!mmatch_r1)
      return MPK_MEMORY;
  mmatch_r2 = (double*) malloc(mmatch_maxstar*sizeof(double));
  if (!mmatch_r2)
      return MPK_MEMORY;
      
  max2 = (mmatch_nstar*(mmatch_nstar-1)*(mmatch_nstar-2))/3+1;
  mmatch_k2  = (double*)malloc(max2*sizeof(double));
  if (!mmatch_k2)
      return MPK_MEMORY;
  mmatch_dev = (double*)malloc(max2*sizeof(double));
  if (!mmatch_dev)
      return MPK_MEMORY;
      
  /* Read reference file */
  sprintf(msg,"Reference file: %s",reffile);
  PrintOut(0, msg);
  res = ReadRef(reffile);
  if (res) 
      return res;
      
	sprintf(msg,"No. of stars: %d",mmatch_c1);
	PrintOut(0, msg);

  	// Check reference file 
	if (mmatch_spfields==0) {
		if (mmatch_c1<mmatch_nstar) {
	        PrintOut(0, "Too few stars in reference file!");
			return MAT_FEW_POINTS_REF;
		}
	} else {
		if (mmatch_c1<1) {
			PrintOut(0, "Too few stars in reference file!");
			return MAT_FEW_POINTS_REF;
		}
	}
  
  	// Normal return 
  	return 0;
}

int match(const char *infile, const char *outfile, int *mstar)
{
	char   msg[MAXLINE];
	int    res;

    *mstar  = 0;
    StInit();
    
    PrintOut(0, "--------------------------------------------");
	sprintf(msg,"Source file                      : %s", infile);
    PrintOut(0, msg);

	// Read input file 
    res = ReadSrc(infile);
    if (res!=0) {
	    StClear();
	    return res;
    }
    
	sprintf(msg,"Number of stars in source file   : %d", mmatch_c2);
	PrintOut(0, msg);
    
    // Find coincidences 
	switch (mmatch_spfields) {
	case 0:
		// Always use standard method
		if (mmatch_c1>=mmatch_nstar && mmatch_c2>=mmatch_nstar) {
			res = Solve(mstar);
		} else {
			PrintOut(0, "Too few stars in source file!");
			res = MAT_FEW_POINTS_SRC;
		}
	    break;
	case 1:
		// Use standard method if we have enough stars
		if (mmatch_c1>=mmatch_nstar && mmatch_c2>=mmatch_nstar) {
		    res = Solve(mstar);
		} else
		if (mmatch_c1>=1 && mmatch_c2>=1) {
		    res = Simple(mstar);
		} else {
			PrintOut(0, "Too few stars in source file!");
			res = MAT_FEW_POINTS_SRC;
		}
		break;
	case 2:
		// Always use method for sparse fields
		if (mmatch_c1>=1 && mmatch_c2>=1) {
			res = Simple(mstar);
		} else {
			PrintOut(0, "Too few stars in source file!");
			res = MAT_FEW_POINTS_SRC;
		}
		break;
    }
    if (res!=0) {
	    StClear();
	    return res;
    }
    
    sprintf(msg,"Number of matched stars          : %d from %d (%.0f%%)", *mstar, mmatch_c2, (100.0**mstar)/mmatch_c2);
    PrintOut(0, msg);
    
    /* Write output file */
    res = MatWrite(outfile);
    if (res!=0) {
	    StClear();
	    return res;
    }

    /* Normal return */
    StClear();
    return res;
}

int match_clean(void)
{
  StClear();
  if (mmatch_idr1) {
      free(mmatch_idr1);
      mmatch_idr1 = NULL;
  }
  if (mmatch_idr2) {
	  free(mmatch_idr2);
	  mmatch_idr2 = NULL;
  }
  if (mmatch_id1) {
      free(mmatch_id1);
      mmatch_id1 = NULL;
  }
  if (mmatch_id2) {
	  free(mmatch_id2);
	  mmatch_id2 = NULL;
  }
  if (mmatch_r1) {
	  free(mmatch_r1);
	  mmatch_r1 = NULL;
  }
  if (mmatch_r2) {
	  free(mmatch_r2);
	  mmatch_r2 = NULL;
  }
  if (mmatch_k2) {
	  free(mmatch_k2);
	  mmatch_k2 = NULL;
  }
  if (mmatch_dev) {
	  free(mmatch_dev);
	  mmatch_dev = NULL;
  }
  return 0;
}
