/**************************************************************

munimatch.c (C-Munipack project)
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#ifdef _WIN32
#include <io.h>
#endif

#include "comfun.h"
#include "match.h"

void PrintHelp()
{
 printf("MUNIMATCH version %s - find cross-references between the files of stars.\n",VERSION);
 printf(PROJECTLINE);
 printf("\nUsage:munimatch [options] [@dirfile] [par=file] ref=file [file1] ...\n\n");
 printf("parameters:\n");
 printf("     @dirfile  file with names of stars catalogues\n");
 printf("               (for only '@', read filenames from standart input)\n");
 printf("     ref=file  name of reference or catalogue file\n\r");
 printf("     par=file  parameter file (default is %s)\n",OPTFILE);
 printf("     file1...  names of source files\n");
 printf("options:\n");
 printf("     -r<file>  reference file (equivalent with 'ref=')\n");
 printf("     -p<file>  parameter file (equivalent with 'par=')\n");
 printf("     -q        quiet mode\n");
 printf("     -h        give this help\n");
 printf("     -v        prints software version\n");
 printf("     -l        display software license\n");
 printf("     -g        creates the example of the parameter file\n");
 printf(WARRANTYLINE);
 printf(HOMEPAGELINE);
 printf(BUGREPORTS);
}

void report(int status, const char *file)
{
 if (cmpack_debug) 
 	fprintf(stderr, "The following error(s) has been occurred: %d\n",status);
 PrintError(PROGNAME, file, 0, status);
}

void error(int status, const char *file)
{
 report(status, file);
 exit(status);
}

void CreatePar()
/* Creates a sample parameter file */
{
 FILE *fp;

 if ((fp=fopen(SMPFILE,"w"))!=NULL) {
   printf("Writing '%s' ...\n\n",SMPFILE);
   fprintf(fp,"RSTARS = %d       # Maximum number of stars used in matching\n",DEFRSTARS);
   fprintf(fp,"ISTARS = %d       # Number of apices of polygons\n",DEFISTARS);
   fprintf(fp,"CLIP   = %.1f     # Clipping factor in sigmas\n",DEFCLIP);
   fclose(fp);
   printf("'%s' is an example of the parameter file. Rename it to '%s' and edit the parameters.\n\n",SMPFILE,OPTFILE);
 }
}

int main(int argc, char **argv)
{
    FILE *f;
    int i, n, stdinput=0, aux, res, level = 0;
    char *par, line[MAXLINE], buf1[MAXLINE], outf[MAXLINE];
	char *namsci = "", *namopt = NULL, *namref = "";

	/* Zpracovani parametru prikazove radky */
	for (i=1;i<argc;i++) {
	   par = argv[i];
	   /* Prints help */
	   if ((strcmp(par,"-h")==0)||(strcmp(par,"--help")==0)) {
	      PrintHelp();
          return 0;
       }
	   /* Prints version */
	   if ((strcmp(par,"-v")==0)||(strcmp(par,"--version")==0)) {
		  PrintVersion(PROGNAME);
          return 0;
       }
       /* Prints license */
       if ((strcmp(par,"-l")==0)||(strcmp(par,"--license")==0)) {
          PrintLicense();
          return 0;
       }
       /* Generate sample parameter file */
       if (strcmp(argv[i],"-g")==0) {
           CreatePar();
           return 0;
       }
       /* Flag for input from stdin or dirfile */
       if (strchr(par,'@')==par) {
	       namsci = malloc(strlen(par));
	       strcpy(namsci,&par[1]);
           stdinput = strlen(namsci)==0;
       }
       /* Reference file */
       if (strstr(par,"ref=")==par) {
	       namref = malloc(strlen(par)-3);
	       strcpy(namref,&par[4]);
       }
       if (strstr(argv[i],"-r")==par) {
	       namref = malloc(strlen(par)-1);
	       strcpy(namref,&par[2]);
       }
       /* Parameter file */
       if (strstr(par,"par=")==par) {
	       namopt = malloc(strlen(par)-3);
	       strcpy(namopt,&par[4]);
       }
       if (strstr(argv[i],"-p")==par) {
	       namopt = malloc(strlen(par)-1);
	       strcpy(namopt,&par[2]);
       }
       /* Quiet/verbose flags */
       if (strcmp(par,"-q")==0 || strcmp(par,"--quiet")==0 || strcmp(par,"--silent")==0) {
           level = MPK_LEVEL_QUIET;
       }
       if (strcmp(par,"--verbose")==0 || strcmp(par,"--debug")==0) {
           level = MPK_LEVEL_DEBUG;
       }
  }

  cmpack_setoutput(PrintMessage, level);
  PrintPrologue(PROGNAME);
  res=match_init(namref,namopt);
  if (res!=0) error(res, NULL);
  n = 0;

  /* image names reads from command line */
  for (i=1;i<argc;i++) {
	  par = argv[i];
      if ((strchr(par,'=')==0)&&(strchr(par,'@')!=par)&&(strchr(par,'-')!=par)) {
#ifdef _WIN32
		intptr_t fh;
		struct _finddata_t fd;
		 if ((fh=_findfirst(par, &fd))!=-1) {
			do {
				if ((fd.attrib & _A_SUBDIR)==0)  {
          			changeext(outf,fd.name,"mat");
          			res=match(fd.name,outf,&aux);
          			if (res!=0) report(res, fd.name);
          			n++;
          		}
          	} while (_findnext(fh, &fd)==0);
			_findclose(fh);
      	 } else {
	      	 report(MPK_CANNOT_OPEN_SRC, par);
      	 }
#else
		changeext(outf,par,"mat");
		res=match(par,outf,&aux);
		if (res!=0) report(res, par);
		n++;
#endif
      }
  }

  /* image names reads from dirfile */
  if (strlen(namsci)>0) {
	  f = fopen(namsci,"r");
	  if (f==NULL) error(MPK_NO_DIRFILE, namsci);
      while(fgets(buf1,MAXLINE,f)!=NULL) {
          sscanf(buf1,"%s",line);
          if (strcmp(line,"")) {
    	      changeext(outf,line,"mat");
	          res=match(line,outf,&aux);
              if (res!=0) report(res, line);
	          n++;
          }
      }
  }

  /* Image names reads from stdin */
  if (stdinput) {
      while(fgets(buf1,MAXLINE,stdin)!=NULL) {
          sscanf(buf1,"%s",line);
          if (strcmp(line,"")) {
    	      changeext(outf,line,"mat");
	          res=match(line,outf,&aux);
              if (res!=0) report(res, line);
	          n++;
          }
      }
  }

  if (n<=0) 
      PrintOut(0, "No input files.");

  /* Normal return */
  res=match_clean();
  if (res!=0) error(res, NULL);

  return 0;
}
