/**************************************************************

phot.c (C-Munipack project)
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

#include "comfun.h"
#include "inifile.h"
#include "phot.h"
#include "attach.h"
#include "find.h"
#include "sky.h"
#include "phwrite.h"
#include "fotometr.h"
#include "sort.h"
#include "phfun.h"

/* Parameters */
double ph_readns;          // Read noise in ADU per frame
double ph_gain;            // Gain in e-/ADU
double ph_fwhm;            // FWHM od object
double ph_thresh;          // Threshold in sigmas
double ph_shrplo;          // Low sharpness cutoff
double ph_shrphi;          // High sharpness cutoff
double ph_rndlo;           // Low roundness cutoff
double ph_rndhi;           // High roundness cutoff
int    ph_fravg;           // Number of frames averaged
int    ph_fravgauto;  	   // Use FR_AVG in the file header
int    ph_frsum;           // Number of frames summed
double ph_minvalue;        // Low pixel cutoff
double ph_maxvalue;        // High pixel cutoff

/* Apertures */
double ph_aper[MAXAP];     // Aperture radii
int    ph_naper;           // Number of apertures used
double ph_is;              // Inner sky radius
double ph_os;              // Outer sky radius

int mphot_init(const char *optfile, const char *apfile)
{
  inifile *f = NULL;
  char msg[MAXLINE];
  int i;

  InitImageInfo();

  /* Read parameter file */
  if (!(f=iniopen(optfile, OPTFILE))) {
	  if (optfile) sprintf(msg,"Cannot read parameter file '%s'", optfile);
	          else sprintf(msg,"Cannot read parameter file '%s'", OPTFILE);
      PrintOut(0, msg);
   	  return MPK_CANNOT_OPEN_PAR;
  }
  ph_readns = inigkyd(f, "READNS" ,   15.0);
  ph_gain   = inigkyd(f, "GAIN"   ,    2.3);
  ph_fwhm   = inigkyd(f, "FWHM"   ,    2.5);
  ph_thresh = inigkyd(f, "THRESH" ,    5.0);
  ph_minvalue  = inigkyd(f, "MINVALUE", 0.0);
  ph_maxvalue  = inigkyd(f, "HIDATA", 65535.0);
  ph_shrplo = inigkyd(f, "LOSHARP",    0.2);
  ph_shrphi = inigkyd(f, "HISHARP",    1.0);
  ph_rndlo  = inigkyd(f, "LOROUND",   -1.0);
  ph_rndhi  = inigkyd(f, "HIROUND",    1.0);
  ph_fravg  = inigkyl(f, "FR_AVG" ,    1);
  ph_fravgauto = inigkyl(f, "FR_AVG" , 0)!=0;
  ph_frsum  = inigkyl(f, "FR_SUM" ,    1);
  iniclos(f);

	sprintf(msg, "READNS: %.2f", ph_readns); 
	PrintOut(1, msg);
	sprintf(msg, "GAIN  : %.2f", ph_gain);  
	PrintOut(1, msg);
    sprintf(msg, "FWHM  : %.2f", ph_fwhm);  
	PrintOut(1, msg);
	sprintf(msg, "THRESH: %.2f", ph_thresh);
	PrintOut(1, msg);
	sprintf(msg, "DATALO: %.2f", ph_minvalue);
	PrintOut(1, msg);
	sprintf(msg, "DATAHI: %.2f", ph_maxvalue);
	PrintOut(1, msg);
	sprintf(msg, "SHRPLO: %.2f", ph_shrplo);
	PrintOut(1, msg);
	sprintf(msg, "SHRPHI: %.2f", ph_shrphi);
	PrintOut(1, msg);
	sprintf(msg, "RNDLO : %.2f", ph_rndlo); 
	PrintOut(1, msg);
	sprintf(msg, "RNDHI : %.2f", ph_rndhi); 
	PrintOut(1, msg);
	if (ph_fravgauto) sprintf(msg, "FRAVG : (Auto)"); 
	             else sprintf(msg, "FRAVG : %d",   ph_fravg); 
	PrintOut(1, msg);
	sprintf(msg, "FRSUM : %d",   ph_frsum); 
	PrintOut(1, msg);

  /* Read aperture file */
  if (!(f=iniopen(apfile, APFILE))) {
       	if (apfile) sprintf(msg,"Cannot read aperture file '%s'",apfile);
		       else sprintf(msg,"Cannot read aperture file '%s'",APFILE);
       	PrintOut(0, msg);
      	return MPK_CANNOT_OPEN_PAR;
  }
  for (i=0; i<MAXAP; i++) {
  	sprintf(msg, "A%d", i+1);
	ph_aper[i] = inigkyd(f, msg, (i==0 ? 2.0 : 0.0));
  }
  ph_is = inigkyd(f, "IS", 20.0);
  ph_os = inigkyd(f, "OS", 30.0);
  iniclos(f);

  /* Compute number of apertures defined */
  ph_naper=0;
  while ((ph_naper<MAXAP)&&(ph_aper[ph_naper]>=1.0)) ph_naper++;
  if (ph_naper<=0) {
    PrintOut(0, "There are no valid apertures defined.");
  	return MPK_INVALID_PAR;
  }
  sprintf(msg,"Number of apertures: %d", ph_naper);
  PrintOut(0, msg);
  for (i=0; i<ph_naper; i++) {
      sprintf(msg, "A[%2d] : %.2f", i+1, ph_aper[i]); 
      PrintOut(1, msg);
  }
  sprintf(msg, "IS    : %.2f", ph_is); 
  PrintOut(1, msg);
  sprintf(msg, "OS    : %.2f", ph_os); 
  PrintOut(1, msg);
  return 0;
}

int mphot(const char *scifile, const char *outfile, double *jd, int *nostar)
{
	FILE *f;
	char  msg[MAXLINE];
	int   res = 0;

	*jd = *nostar = 0;
	PrintOut(0, "--------------------------------------------");
	sprintf(msg,"Image file   : %s",scifile);
    PrintOut(0, msg);

    /* Check source file */
    if (!scifile) {
        PrintOut(0, "Missing name of the image file.");
 	    return MPK_CANNOT_OPEN_SRC;
    }
    if (strlen(scifile)==0) {
        PrintOut(0, "Missing name of the image file.");
 	    return MPK_CANNOT_OPEN_SRC;
    }
    f = fopen(scifile,"r");
    if (f==NULL) {
 	    PrintOut(1, "Image file file not found.");
 	    return MPK_CANNOT_OPEN_SRC;
    }
    fclose(f);

    InitStarList();

    /* Attach the image */
    res = Attach(scifile);
    if (res==0) {
	    *jd = ph_image.jd;
        sprintf(msg,"JD (UT)      : %.6f", ph_image.jd);
        PrintOut(0, msg);
        if (ph_fravg>1) {
        	sprintf(msg,"Frames avgd  : %d", ph_fravg);
        	PrintOut(0, msg);
        }
    }

    /* Sky */
    if (res==0) {
        res = Sky();
        if (res==0) {
	        sprintf(msg,"Sky value    : %.1f +- %.1f ADU", ph_image.skymod, ph_image.skysig);
	        PrintOut(0, msg);
        }
    }

    /* Find stars */
    if (res==0) {
        res = Find();
        if (res==0) {
	        sprintf(msg,"Rel. error   : %.2f", ph_image.relerr);
	        PrintOut(0, msg);
            sprintf(msg,"Stars found  : %d", ph_nstar);
            PrintOut(0, msg);
        }
        *nostar = ph_nstar;
    }

    /* Aperture photometry */
    if (res==0) {
	    res = PhotSB();
        if ((res==0)&&(ph_image.maglim<99.0)&&(ph_image.magsq<9.0)) {
            sprintf(msg,"Mag. limit   : %.2f +- %.2f per star in aperture #1.", ph_image.maglim, ph_image.magsq);
            PrintOut(0, msg);
            sprintf(msg,"Mean FWHM    : %.2f +- %.2f pixels.", ph_image.fwhm_med, ph_image.fwhm_err);
            PrintOut(0, msg);
        }
    }

    /* Sort of file according to key(s) */
    if (res==0) {
	    res = Sorter();
    }

    /* Write output to photometry file */
    if (res==0) {
	    res = SrtWrite(outfile);
	    if (res==0) {
		    sprintf(msg,"Output file  : %s",outfile);
            PrintOut(0, msg);
        }
    }

    /* Normal return */
    ClearStarList();
    return res;
}

int mphot_clean(void)
{
  ClearStarList();
  ClearImageInfo();
  return 0;
}
