/**************************************************************

timebat.c (C-Munipack project)
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <fitsio.h>

#ifdef _WIN32
#include <io.h>
#endif

#include "comfun.h"
#include "tcor.h"

void PrintHelp()
{
 printf("TIMEBAT version %s - utility for the time correction.\n",VERSION);
 printf(PROJECTLINE);
 printf("\nUsage:timebat [options] [@dirfile] [tcor=seconds] [mask=?] [image1] ...\n\n");
 printf("parameters:\n");
 printf("         @dirfile= file with image names for time correction,\n");
 printf("                   (for only '@', read image names from standart input)\n");
 printf("         tcor=     amount of time in seconds (can be real number)\n");
 printf("         image1... names of simple images to time correction\n");
 printf("         mask=     output mask (default = %s)\n",DNAME);
 printf("                   give 'mask=.' for overwrite input files (carefully!)\n");
 printf("options:\n");
 printf("        -t<sec>    amount of time (equivalent with 'tcor=')\n");
 printf("        -o<mask>   output mask (equivalent with 'mask=')\n");
 printf("        -c<number> counter initial value (default = 1)\n");
 printf("        -q         quiet mode\n");
 printf("        -h         give this help\n");
 printf("        -v         prints software version\n");
 printf("        -l         display software license\n");
 printf(WARRANTYLINE);
 printf(HOMEPAGELINE);
 printf(BUGREPORTS);
}

void report(int status, const char *file)
{
 if (cmpack_debug)
 	fprintf(stderr, "The following error(s) has been occurred: %d\n",status);
 if (status<1000)
 	ffrprt(stderr, status);
 else
 	PrintError(PROGNAME, file, 0, status);
}

void error(int status, const char *file)
{
 report(status, file);
 exit(status);
}

int main(int argc, char **argv)
{
	int i, icit, res, icit0 = 1, stdinput = 0, level = 0;
	double seconds = 0.0;
	char *namsci = "", *mask = DNAME, *par;
	char format[MAXLINE], outf[MAXLINE], line[MAXLINE], buf1[MAXLINE];
	FILE *f;

	/* Zpracovani parametru prikazove radky */
	for (i=1;i<argc;i++) {
	   par = argv[i];
	   /* Prints help */
	   if ((strcmp(par,"-h")==0)||(strcmp(par,"--help")==0)) {
		  PrintHelp();
          return 0;
       }
	   /* Prints version */
	   if ((strcmp(par,"-v")==0)||(strcmp(par,"--version")==0)) {
		  PrintVersion(PROGNAME);
          return 0;
       }
       /* Prints license */
       if ((strcmp(par,"-l")==0)||(strcmp(par,"--license")==0)) {
          PrintLicense();
          return 0;
       }
       /* Flag for input from stdin or dirfile */
       if (strchr(par,'@')==par) {
	       namsci = malloc(strlen(par));
	       strcpy(namsci,&par[1]);
           stdinput = strlen(namsci)==0;
       }
       /* File with flat frame */
       if (strstr(par,"tcor=")==par) {
	       seconds = atof(&par[5]);
       }
       if (strstr(par,"-t")==par) {
	       seconds = atof(&par[2]);
       }
       /* Mask of output files */
       if (strstr(par,"mask=")==par) {
	       mask = malloc(strlen(par)-4);
	       strcpy(mask,&par[5]);
       }
       if (strstr(par,"-o")==par) {
	       mask = malloc(strlen(par)-1);
	       strcpy(mask,&par[2]);
       }
       /* Initial counter value */
       if (strstr(par,"-c")==par) {
	       icit0 = atol(&par[2]);
       }
       /* Quiet/verbose flags */
       if (strcmp(par,"-q")==0 || strcmp(par,"--quiet")==0 || strcmp(par,"--silent")==0) {
           level = MPK_LEVEL_QUIET;
       }
       if (strcmp(par,"--verbose")==0 || strcmp(par,"--debug")==0) {
           level = MPK_LEVEL_DEBUG;
       }
  }

  cmpack_setoutput(PrintMessage, level);
  PrintPrologue(PROGNAME);
  res=tcor_init(seconds);
  if (res!=0) error(res, NULL);

  /* prepare format string for generating output file names */
  if (strcmp(mask,".")==0) {
     mask = NULL;            /* dot "." as output mask */
  } else {
     MaskToFormat(format,mask);
  }

  /* counter initialization */
  icit = icit0;

  /* image names reads from command line */
  for (i=1;i<argc;i++) {
	par = argv[i];
	if ((strchr(par,'=')==0)&&(strchr(par,'@')!=par)&&(strchr(par,'-')!=par)) {
#ifdef _WIN32
		intptr_t fh;
		struct _finddata_t fd;
		if ((fh=_findfirst(par, &fd))!=-1) {
			do {
				if ((fd.attrib & _A_SUBDIR)==0)  {
      					if (mask) {
	      					sprintf(outf,format,icit);
     	  					res=tcor(fd.name,outf,NULL,NULL);
       					} else {
         					res=tcor(fd.name,NULL,NULL,NULL);
      					}
  					if (res!=0) report(res, fd.name);
					icit++;
				}
          		} while (_findnext(fh, &fd)==0);
			_findclose(fh);
      	 } else {
	      	 report(MPK_CANNOT_OPEN_SRC, par);
      	 }
#else
		if (mask) {
			sprintf(outf,format,icit);
			res=tcor(par,outf,NULL,NULL);
		} else {
			res=tcor(par,NULL,NULL,NULL);
		}
		if (res!=0) report(res, par);
		icit++;
#endif
      }
  }

  /* image names reads from dirfile */
  if (strlen(namsci)>0) {
	  f = fopen(namsci,"r");
	  if (f==NULL) error(MPK_NO_DIRFILE, namsci);
      while(fgets(buf1,MAXLINE,f)!=NULL) {
          sscanf(buf1,"%s",line);
          if (strcmp(line,"")) {
	      	  if (mask) {
		      	  sprintf(outf,format,icit);
	      	  	  res=tcor(line,outf,NULL,NULL);
              } else {
	              res=tcor(line,NULL,NULL,NULL);
              }
              if (res!=0) report(res, line);
              icit++;
          }
      }
  }

  /* Image names reads from stdin */
  if (stdinput) {
      while(fgets(buf1,MAXLINE,stdin)!=NULL) {
          sscanf(buf1,"%s",line);
          if (strcmp(line,"")) {
	      	  if (mask) {
		      	  sprintf(outf,format,icit);
	      	  	  res=tcor(line,outf,NULL,NULL);
              } else {
	              res=tcor(line,NULL,NULL,NULL);
              }
              if (res!=0) report(res, line);
              icit++;
          }
      }
  }

  if (icit==icit0)
      PrintOut(0, "No input files.");

  res=tcor_clean();
  if (res!=0) error(res, NULL);

#ifdef _WIN32
  //FitsFreeLibrary();
#endif
  return 0;
}
